# -*- coding: UTF-8 -*-
#
#  DiskInfo.py : Detect and get informations about block devices
#  Copyright (C) 2007 Mertens Florent <flomertens@gmail.com>
#  Updated 2021 for MX Linux Project by team member Nite Coder
#  Maintenance of project assumed by MX Linux with permission from original author.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import os
import sys
import glob
import logging
import subprocess
from subprocess import *

from DiskManager.Fstab.Fstabconfig import *
from DiskManager.Fstab.FstabError import *
from DiskManager.Fstab import FstabData

REQUIRED = ["DEV", "DEVICE", "DEV_NUM", "FS_TYPE"]
DEFAULT_BACKEND = "ToolsBackend"


def get_diskinfo_backend(backend = "auto") :
    ''' get_diskinfo_backend([backend]) -> return the best backend available, 
                                           or the backend specify by backend.\n
        ex : info = get_diskinfo_backend()() '''

    backends = []
    path = os.path.dirname(__file__)
    if not path :
        path = os.getcwd()
    sys.path.append(path)
    names = [ os.path.basename(path)[:-3] for path in os.listdir(path)
              if path.endswith("Backend.py") ]
    for name in names:
        module = __import__( name )
        for attr in dir(module):
            obj = getattr(module, attr)
            if hasattr(obj, "_backend_score") :
                backends.append(obj)
    if len(backends) == 0 :
        logging.warning("No DiskInfo backend found. Trying the default one...")
        module = __import__( DEFAULT_BACKEND )
        return getattr(module, DEFAULT_BACKEND)
    if not backend == "auto" :
        backend = backend[0].upper() + backend[1:] + "Backend"
        if not backend in [ k.__name__ for k in backends ] :
            logging.warning("%s is not available. Using auto backend." % backend)
        elif getattr(module, backend )()._backend_score() :
            return getattr(module, backend)
        else :
            logging.warning("%s is not currently working. Using auto backend." % backend)
    scores = [ k()._backend_score() for k in backends ]
    return backends[scores.index(max(scores))]


class DiskInfoBase(list) :
    ''' Base class for all DiskInfo backend. This class implememt all the infrastructure
        to manage DiskInfo class, and leave only to backends the works to get sensible
        informations about a device. '''

    def __init__(self) :

        self._loaded = False
        self._loading = False
        self._driver_db = {}
        
    def _backend_score(self) :
        ''' Return the reliability score of the backend (0-100). You must override
            this method when creating a backend. You might probably make this score
            dependant of the versions of the tools you use. Keep also in mind that
            the score of the default backend (ToolsBackend) is set to 50 '''
             
        return 0
        
    def _get_devices(self) :
        ''' This function try to get all block devices from /proc/partitions. You might or not
            override this method, dependending on how reliable you think it is. '''
    
        fd = open("/proc/partitions")
        devices = fd.readlines()[2:]
        fd.close()
        return [x for x in devices if all(y not in x for y in ["fd", "loop", "sr", "ram"])]
        
    def load_database(self, reload = False) :
        ''' x.load_database([reload]) -> load the database.\n
            At DiskInfo creation, the database is not created. You have to call load_database
            to create it. But you should not need to explicitly doing it, since all other methods
            that use the database should do it. In fact this method is keeped public only
            for it's reload option : when calling load_database with reload = True, the
            database is reloaded '''
            
        if (self._loaded or self._loading) and not reload : 
            return
        logging.debug("Starting DiskInfo database creation...")
        self._loading = True
        del self[:]
        devices = self._get_devices()
        for i in range(len(devices)) :
            self.append({})
            self[i] = self._get_device_info(devices[i])
            self[i] = self._get_common_device_info(self[i])
            logging.debug("-"*10)
        self._load_reverse_database()
        self._loaded = True
        self._loading = False
        logging.debug("End DiskInfo database creation.")
        
    def _get_device_info(self, device) :
        ''' This is the method that all backends must override. It should returns a dict of
            information about device. Informations are :
            "DEV"       : simple device name, like sda1 or dm-0. (required)
            "DEVICE"    : the device path, like /dev/sda1, or /dev/mapper/*. (required)
            "FS_UUID"   : the UUID of the device (required)
            "FS_LABEL"  : the label of the device (required if available)
            "FS_TYPE"   : the type of the device, like ext3. (required)
            "FS_USAGE"  : the usage of the fs, like filesystem, or crypto.
                          Devices with FS_USAGE != filesystem are ignored. (required)
            "REMOVABLE" : set to True if the device is from a removable drive.
                          Removable devices are ignored. (required)
            "MINOR"     : the minor number of the device (required)
            "MAJOR"     : the major number of the device (required)
            "SYSFS_PATH": the sysfs path (recommanded)
            "SIZE"      : the size of the device (recommanded)
            "FS_VERSION": the version of the type (recommanded)
            "MODEL", "SERIAL_SHORT", "BUS", "VENDOR", "SERIAL", "TYPE", "REVISION" :
            some informations about the device, and its drive. (optionnal) '''
    
        #Impement your backend here
        return device
        
    def _get_common_device_info(self, dev) :
        ''' This method complete device informations with common attributes that should not
            be the job of a backend. Also set here what we ignore or not. '''

        # Set FS_DRIVERS
        if "FS_TYPE" in dev and "FS_USAGE" in dev and dev["FS_USAGE"] == "filesystem" :
            dev["FS_DRIVERS"] = self.get_drivers(dev["FS_TYPE"])
            
        # Set FS_LABEL_SAFE
        if "FS_LABEL" in dev and "FS_LABEL_SAFE" not in dev :
            label = dev["FS_LABEL"]
            label = label.replace("/", "")
            for char in FstabData.special_char :
                label = label.replace(char, "_")
            if label :
                logging.debug("-> Set   FS_LABEL_SAFE : " + label)
                dev["FS_LABEL_SAFE"] = label
                
        # Set DEV_NUM
        if "MINOR" in dev and "MAJOR" in dev :
            dev["DEV_NUM"] = os.makedev(int(dev["MAJOR"]), int(dev["MINOR"]))

        # Ignore every dev if FS_USAGE != filesystem or if REMOVABLE
        # also ignore dev that don't have all REQUIRED attributes
        dev["IGNORE"] = False
        if "FS_USAGE" not in dev :
            dev["IGNORE"] = True
        elif not dev["FS_USAGE"] == "filesystem" :
            dev["IGNORE"] = True
        if not min(list(map(lambda x: x in dev, REQUIRED))) :
            dev["IGNORE"] = True
        if dev["REMOVABLE"] :
            dev["IGNORE"] = True

        # Ignore older entry with the same DEVICE
        if "DEVICE" in dev :
            for device in self.search(dev["DEVICE"], keys = ["DEVICE"]) :
                if not self[device]["DEV"] == dev["DEV"] :
                    logging.debug("W: " + "Ignore duplicate entry : " + self[device]["DEV"] \
                        + " -> " + self[device]["DEVICE"])
                    self[device]["IGNORE"] = True

        return dev
        
    def _load_reverse_database(self) :
    
        self._reverse_database = {}
        for i in range(len(self)) :
            if "DEV_NUM" in self[i] :
                self._reverse_database["DEV_NUM=%s" % self[i]["DEV_NUM"]] = i
            if "FS_UUID" in self[i] :
                self._reverse_database["UUID=%s" % self[i]["FS_UUID"]] = i
            if "FS_LABEL" in self[i] :
                self._reverse_database["LABEL=%s" % self[i]["FS_LABEL"]] = i
        
    def get_drivers(self, type, reload = False) :
        ''' x.get_drivers(type, [reload]) -> return a dict of available driver for this type.\n
            Return dict is of the type : {primary : [[name1, description1, fsck1], ...], \
                secondary : [[name2, description2, fsck2], ...], \
                all : {name1 : [name1, description1, fsck1], name2 : [name2, description2, fsck2], ...}}
            fsck = 1 if it exists an fsck for this driver, otherwise it is 0.
            drivers are checked in /proc/filesystem, modprobe -l, and in /sbin/mount.*.
            To avoid having to do the work again and again, result is cached, and so
            you ll need to set reload to True to have up to date results. '''
    
        if type in self._driver_db and not reload :
            return self._driver_db["type"]

        self._driver_db["type"] = { "primary" : [], "secondary" : [], "all" : {} }

        # special ntfs handling
        if type == "ntfs":
            for special in glob.glob("/sbin/mount.%s*" % type) :
                if os.path.isfile(special) :
                    special = special.split("mount.")[-1]
                    self._driver_db["type"]["primary"].append([special, \
                        FstabData.special_driver.get(special, FstabData.special_driver["__unknow__"])])

            if "ntfs3" in open("/proc/filesystems").read() or self._check_module("ntfs3") :
                self._driver_db["type"]["primary"].append(
                        ["ntfs3", FstabData.special_driver.get("ntfs3", FstabData.special_driver["__unknow__"])])

        # non-ntfs driver detection
        else:
            if type in open("/proc/filesystems").read() or self._check_module(type) :
                self._driver_db["type"]["primary"].append([type, "Default driver"])
            for special in glob.glob("/sbin/mount.%s*" % type) :
                if os.path.isfile(special) :
                    special = special.split("mount.")[-1]
                    self._driver_db["type"]["primary"].append([special, \
                        FstabData.special_driver.get(special, FstabData.special_driver["__unknow__"])])

        if "__all__" in FstabData.secondary_driver :
            self._driver_db["type"]["secondary"].append([ \
                FstabData.secondary_driver["__all__"], "Secondary Driver"])
        if type in FstabData.secondary_driver :
            self._driver_db["type"]["secondary"].append([ \
                FstabData.secondary_driver[type], "Secondary Driver"])
        for driver in self._driver_db["type"]["primary"] + self._driver_db["type"]["secondary"] :
            driver.append(int(os.path.isfile("/sbin/fsck.%s" % driver[0])))
            self._driver_db["type"]["all"][driver[0]] = driver
        return self._driver_db["type"]

    def _check_module(self, module) :
        if module == "ntfs3":
            try:
                cmd = f"{MODPROBE} -b -i {module} 2>/dev/null 1>&2 && lsmod | grep -o {module} 2>/dev/null"
                result = run(cmd, shell=True, capture_output=True, text=True)
                return module in result.stdout
            except CalledProcessError:
                return False
        else:
            try:
                cmd = [MODPROBE, "-n", "-i", module]
                run(cmd, check=True, stderr=DEVNULL, stdout=DEVNULL)
                return True
            except CalledProcessError:
                return False

    def __getitem__(self, item) :
        
        self.load_database()
        if type(item) == int :
            if item < len(self) :
                return list.__getitem__(self, item)
            else :
                raise NotInDatabase("Index %i out of range" % item)
        else :
            try :
                return list.__getitem__(self, self.search(item)[0])
            except :
                raise NotInDatabase("Can't find %s in the database" % item)
            
    def list(self, col = "DEVICE", ignored = True, keep_index = False) :
        ''' x.list([col], [ignored], [keep_index]) -> List all values of attribute col.
                                                      Default to "DEVICE"\n
            If ignored is set to False, don't list device with IGNORE=True. Default to True.
            If keep_index is set to True and ignored to False, all ignored device result to
            an empty string, to keep the index with the database. Default to False. ''' 

        self.load_database()
        result = []
        for k in self :
            if not ignored and k["IGNORE"] :
                if keep_index :
                    result.append("")
                continue
            if col in k :
                result.append(k[col])
            else :
                result.append("")
        return result
        
    def get(self, item, attribute) :
        ''' x.get(item, attribute) -> return attribute of item '''
    
        self.load_database()
        try :
            return self[item][attribute]
        except :
            return "None"
            
    def search_reverse(self, pattern, ignored = True) :
        ''' x.search_reverse(pattern, [ignored]) -> search for pattern in the reverse database\n
            This method is faster than the search method, but not as flexible, since pattern is
            searched in fixed keys : "DEV_NUM", "FS_UUID", "FS_LABEL"
            Set ignored to False, to not return device with IGNORE=True. Default to True '''

        self.load_database()
        if self._loaded and pattern in self._reverse_database :
            result = self._reverse_database[pattern]
            if ignored or not self[result]["IGNORE"] :
                return result
        return None

    def search_device(self, entry, ignored = True) :
        ''' x.search_device(entry, ignored = True) -> search entry in the reverse database\n
            search_device is a convenient method that will check if device is of
            the type dev_file, "UUID=", "LABEL=" and will search in the reverse database
            the appropriate value.
            Set ignored to False, to not return device with IGNORE=True. Default to True '''

        if os.path.exists(entry) :
            entry = "DEV_NUM=%s" % os.stat(os.path.realpath(entry)).st_rdev
        return self.search_reverse(entry, ignored)

    def search(self, pattern, keys = ["DEV", "DEVICE"], ignored = True) :
        ''' x.search(pattern, [list], [keys], [ignored]) -> search for pattren in each
                                                            keys of each Entry of x\n
            Default keys are : ["DEV", "DEVICE"]
            Set ignored to False, to not return device with IGNORE=True. Default to True '''

        self.load_database()
        result = []
        for col in keys :
            i = 0
            for value in self.list(col, ignored, keep_index = True) :
                if value == pattern and i not in result :
                    result.append(i)
                if col == "FS_UUID" :
                    if "UUID=" + value == pattern and i not in result :
                        result.append(i)
                if col == "FS_LABEL" :
                    if "LABEL=" + value == pattern and i not in result :
                        result.append(i)
                i = i + 1
        return result

    def export(self, device) :
        ''' x.export(device) -> query database for device, and return a string of printable
                                informations about device.\n
            If device is set to "all", query database for all devices '''
    
        self.load_database()
        result = ""
        if device == "all" :
            result += "Query database for all devices :\n\n"
            for i in range(len(self)) :
                result += "Info for " + self[i]["DEV"] + " :\n"
                result += "\n".join(["-> %s=%s" % (k, v) for k, v in list(self[i].items())])
                result += "\n\n"
        else :
            result += "Query database for " + device + " :\n\n"
            if self.search(device) :
                result += "Info for " + self[device]["DEV"] + " :\n"
                result += "\n".join(["-> %s=%s" % (k, v) for k, v in list(self[device].items())])
                result += "\n\n"
            else :
                result += device + " not in the database"
        return result



# -*- coding: UTF-8 -*-
#
#  Dialogs.py : Various dialogs for DiskManager
#  Copyright (C) 2007 Mertens Florent <flomertens@gmail.com>
#  Updated 2021 for MX Linux Project by team member Nite Coder
#  Maintenance of project assumed by MX Linux with permission from original author.
#
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import re

import time
from xml.sax.saxutils import escape as escape_mkup

from gi.repository import Gtk
from gi.repository import GdkPixbuf, GObject, Pango

import gettext
#from gettext import gettext as _

from .SimpleGladeApp import SimpleGladeApp

from .config import *
from .Config  import *
from .Fstab.FstabHandler import *
from .Utility import *

import logging


class EditPartition(SimpleGladeApp) :
    ''' Class to manage the Edit/Add device dialog '''
        
    def __init__(self, disk, entry, parent = None) :
        
        SimpleGladeApp.__init__(self, GLADEFILE, "dialog_edit", domain = "disk-manager")
        self.dialog_edit.set_title("")
        if parent : 
            self.dialog_edit.set_transient_for(parent)

        self.disk = disk
        self.entry = entry
        self.disk.set_mode("virtual")
        
        self.default_button.connect("clicked", self.on_default_clicked)
        self.browser_button.connect("clicked", self.on_browser_clicked)
        self.driver_box.connect("changed", self.on_driver_changed)
        self.path.connect("activate", self.on_apply)
        self.options3.connect("activate", self.on_apply)
        self.apply_button2.connect("clicked", self.on_apply)
        self.dialog_edit.connect("delete-event", self.on_quit)
        self.cancel_button2.connect("clicked", self.on_quit)
        
        # Gtk.Tooltips().set_tip(self.default_button, _("Return options to defaults"))
        # Gtk.Tooltips().set_tip(self.browser_button, _("Select a directory"))
         
        # deduplicate mount options
        self.entry["FSTAB_OPTION"] = ','.join(dict.fromkeys(self.entry["FSTAB_OPTION"].split(',')))
        self.path.set_text(self.entry["FSTAB_PATH"])
        self.options3.set_text(self.entry["FSTAB_OPTION"])
        self.vbox_driver.hide()
        if "FS_DRIVERS" in self.entry :
            self.drivers = self.entry["FS_DRIVERS"]
        else :
            self.drivers = {"primary" : [], "all" : {}, "secondary" : []}
        drivers_list = self.drivers["primary"][:]
        driver = self.entry["FSTAB_TYPE"]
        if not driver in [ k[0] for k in drivers_list ] :
            if driver in self.drivers["all"] :
                drivers_list.append(self.drivers["all"][driver])
            elif driver == "ntfs3" :
                drivers_list.append([driver, "Read-write kernel driver", 0])
            else :
                drivers_list.append([driver, "Unknow driver", 0])
        if len(drivers_list) > 1 or not self.drivers["primary"] :
            self.vbox_driver.show()
            self.driver_box.clear()
            self.liststore = Gtk.ListStore(str)
            renderer = Gtk.CellRendererText()
            self.driver_box.set_model(self.liststore)
            self.driver_box.pack_start(renderer, True)
            self.driver_box.add_attribute(renderer, 'text', 0)
            current_driver_focus = 0
            for i in range(len(drivers_list)) :
                self.driver_box.append_text("%s (%s)" % \
                    (drivers_list[i][0], drivers_list[i][1]))
                if driver == drivers_list[i][0] :
                    current_driver_focus = i
            self.driver_box.set_active(current_driver_focus)
        self.set_driver()
        self.dialog_edit.set_focus(self.path)
        if entry.get_is_system() :
            self.path.set_sensitive(False)
            self.browser_button.set_sensitive(False)
        self.dialog_edit.show()
        
    def on_default_clicked(self, button) :
        ''' Load default options '''

        self.options3.set_text(self.entry.defaultopt())

    def set_driver(self) :
        fsk = 0
        if self.entry["FS_TYPE"] in ('ntfs', 'ntfs3', 'exfat') :
            self.driver_warning_box.hide()
            self.driver_error_box.hide()
        if self.drivers["primary"] :
            if self.entry["FSTAB_TYPE"] in self.drivers["all"] :
                self.driver_warning_box.hide()
                fsk = self.drivers["all"][self.entry["FSTAB_TYPE"]][2]
            else :
                if self.entry["FS_TYPE"] not in ('ntfs', 'ntfs3', 'exfat') :
                    self.driver_warning_box.show()
                    fsk = 0
        else :
            self.driver_box.set_sensitive(False)
            self.driver_error_box.show()
            self.driver_error_label.set_label( \
               _("No driver is available for this type of filesystem : '%s'") \
               % self.entry["FS_TYPE"])
            fsk = 0
        self.check_button.set_active(bool(int(self.entry["FSTAB_PASO"])))
        self.check_button.set_sensitive(bool(self.check_button.get_active() or fsk))

    def on_browser_clicked(self, button) :
    
        """  not used:
        gladexml =  Gtk.glade.XML(GLADEFILE, "dialog_mount_point")
        dialog = gladexml.get_widget("dialog_mount_point")

        builder = Gtk.Builder()
        builder.add_from_file(GLADEFILE)

        dialog = builder.get_object("dialog_mount_point")

        dialog.set_transient_for(self.dialog_edit)
        dialog.select_filename(self.path.get_text())
        for shortcut in (get_user("dir"), "/media", "/") :
            try :
                dialog.add_shortcut_folder(shortcut)
            except gobject.GError : 
                pass
        ret = dialog.run()
        print(ret)
        print(dialog.get_filename())
        
        if ret == 1 :
            self.path.set_text(dialog.get_filename())
        dialog.hide()

        dialog = Gtk.FileChooserDialog(title="Please choose a folder", action=Gtk.FileChooserAction.SELECT_FOLDER)
        """
        
        dialog = Gtk.FileChooserDialog(title=_("Select a directory"), action=Gtk.FileChooserAction.SELECT_FOLDER)

        gtk_select = "_Select"
        
        dialog.add_buttons(
            Gtk.STOCK_CANCEL, 
            Gtk.ResponseType.CANCEL, 
            gettext.translation('gtk30').gettext(gtk_select), 
            Gtk.ResponseType.OK
        )
       #dialog.add_buttons(
       #     Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, "Select", Gtk.ResponseType.OK
       # )

        logging.debug("Mountpoint: " + self.path.get_text())
        user_mp_dir = os.path.dirname(os.path.realpath(self.path.get_text()))

        default_mp_path = "/media"
        if os.path.isdir(user_mp_dir):
             mp_path = user_mp_dir
        else:
            mp_path = default_mp_path
        
        dialog.set_current_folder(mp_path)
        dialog.set_default_size(800, 400)

        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            print("Select clicked")
            print("Folder selected: " + dialog.get_filename())
            self.path.set_text(dialog.get_filename())
        elif response == Gtk.ResponseType.CANCEL:
            print("Cancel clicked")

        dialog.destroy()

    def on_driver_changed(self, button) :
    
        current = self.driver_box.get_active_text().split()[0]
        if self.entry["DEVICE"] and not self.entry["FSTAB_TYPE"] == current :
            self.disk.set(self.entry, type=current)
            # deduplicate mount options
            self.entry["FSTAB_OPTION"] = ','.join(dict.fromkeys(self.entry["FSTAB_OPTION"].split(',')))

            self.options3.set_text(self.entry["FSTAB_OPTION"])
            self.set_driver()

    def on_apply(self, widget) :
    
        if not self.check_new_options() and not self.check_new_path() :
            self.disk.set(self.entry, path=self.path.get_text(), option=self.options3.get_text(), \
                paso=self.check_button.get_active())
            self.disk.apply_now()
            self.on_quit(widget)
    
    def on_quit(self, widget, event = None) :
    
        self.dialog_edit.hide()
        self.disk.set_previous_mode()
        
    def check_new_path(self) :
        ''' Check that the path is correct. Need to success to the check_path() test. '''
        
        path = self.path.get_text()
        if len(path) == 0 :
            self.path.set_text(self.entry["FSTAB_PATH"])
            return 1
        if not path[0] == "/" :
            path = "/media/" + path
            self.path.set_text(path)
        if check_path(path, fstab = self.disk, entry = self.entry) :
            return 0
        dialog("warning", _("Wrong mount point"), \
            _("<i>%s</i> is currently in use.\n" \
            "You should try another one.") % escape_mkup(path), parent = self.dialog_edit)
        self.dialog_edit.set_focus(self.path)
        return 1

    def check_new_options(self) :
        ''' Check that options are of the type : jsjk,skks,skks '''
    
        options = self.options3.get_text()
        if re.search("^([a-zA-Z0-9=_.@/-]+,)*[a-zA-Z0-9=_.@/-]+$", options) :
            return 0     
        dialog("warning", _("Options formatting error"), \
                _("<i>%s</i> is not formatted correctly.\n" \
                "Options should be separated by comma (,)") % escape_mkup(options), \
                parent = self.dialog_edit)
        self.dialog_edit.set_focus(self.options3)
        return 1

""" 
# (not used)

class HistoryDialog(SimpleGladeApp) :
    ''' Class the manage the History dialog '''
             
    def __init__(self, disk, parent = None):
    
        SimpleGladeApp.__init__(self, GLADEFILE, "dialog_history", domain = "disk-manager")
        self.dialog_history.set_title("")
        if parent: 
            self.dialog_history.set_transient_for(parent)
        
        self.disk = disk
        
        self.apply_button.connect("clicked", self.on_apply_clicked)
        self.see_button.connect("clicked", self.on_see_clicked)
        self.clean_button.connect("clicked", self.on_clean_clicked)
        self.close_button.connect("clicked", self.on_quit)
        self.dialog_history.connect("delete-event", self.on_quit)
        
        renderer = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn("historic", renderer, text=0, editable=1)
        self.treeview.append_column(column)
        
        self.tree_store = Gtk.ListStore(str, bool)
        self.treeview.set_model(self.tree_store)
        self.update_dialog()
        
    def update_dialog(self) :
        
        self.tree_store.clear()
        self.versions = self.disk.get_listlog()
        self.versions.sort()
        self.versions.reverse()
        if "Original" in self.versions :
            self.versions.remove("Original")
            self.versions.append("Original")
        for value in self.versions :
            if not value == "Original" : 
                value = time.strftime("%d %b %Y %H:%M:%S", time.localtime(float(value)))
            self.tree_store.append((value, False))
        self.treeview.connect("cursor_changed", self.on_cursor_changed)
        first = self.tree_store.get_path(self.tree_store.get_iter_first())
        self.treeview.set_cursor(first)
              
    def on_cursor_changed(self, treeview) :
    
        self.apply_button.set_sensitive(False)
        version = self.versions[self.treeview.get_cursor()[0][0]]
        changes = self.disk.get_changes_current_to(version)
        buf = Gtk.TextBuffer()
        tag = buf.create_tag(weight=Pango.Weight.BOLD)
        buf.insert_with_tags(buf.get_start_iter(), "Commit:\n", tag)
        buf.insert_at_cursor(self.disk.get_logcommit(version))
        if changes :
            iter = buf.get_end_iter()
            buf.insert_with_tags(iter, "\nChanges between this version and current one:\n", tag)
            buf.insert_at_cursor("\n".join(changes))
            self.apply_button.set_sensitive(True)
        self.textview.set_buffer(buf)
        
    def on_apply_clicked(self, button) :
    
        version = self.versions[self.treeview.get_cursor()[0][0]]
        ret = dialog("question", _("Reverting to an older version?"), \
            [_("This will apply the following changes:"), _("Do you want to continue?")],
            "\n".join(self.disk.get_changes_current_to(version)), parent = self.dialog_history)
        if ret[0] == Gtk.ResponseType.YES :
            self.disk.revert_to(version)
        self.on_cursor_changed(self.treeview)
        
    def on_see_clicked(self, button) :
    
        version = self.versions[self.treeview.get_cursor()[0][0]]
        gladexml =  Gtk.glade.XML(GLADEFILE, "dialog_fstab")
        dialog = gladexml.get_widget("dialog_fstab")
        dialog.set_title("")
        dialog.set_transient_for(self.dialog_history)
        textview = gladexml.get_widget("textview_fstab")
        buf = Gtk.TextBuffer()
        if not version == "Original" :
            buf.set_text(FstabData.header)
        buf.insert_at_cursor(self.disk.get_logfile(version))
        textview.set_buffer(buf)
        dialog.run()
        dialog.hide()
        
    def on_clean_clicked(self, button) :
    
        ret = dialog("question", _("Emptying History?"), \
            _("Do you want to empty the History?"), parent = self.dialog_history)
        if ret[0] == Gtk.ResponseType.YES :
            self.disk.cleanlog()
            self.update_dialog()
            
    def on_quit(self, button, event = None) :
    
        self.dialog_history.hide()
"""
        
        
class InfoDialog(SimpleGladeApp) :

    def __init__(self, entry, disk, parent = None) :

        SimpleGladeApp.__init__(self, GLADEFILE, "dialog_info", domain = "disk-manager")
        if parent :
            self.dialog_info.set_transient_for(parent)
        self.entry = entry
        self.disk = disk
        self.infobuttonclose.connect("clicked", self.on_quit)

    def update_dial(self) :
        gtk_information = "Information"
        self.dialog_info.set_title(self.entry["DEV"] + " " + gettext.translation('gtk30').gettext(gtk_information))
        self.size_label.set_label(size_renderer(self.entry.get_size()))
        if self.entry.get_is_mounted() :
            self.free_label.set_label(size_renderer(self.entry.get_free_size()))
            self.avail_label.set_label(size_renderer(self.entry.get_available_size()))
            self.used_label.set_label(size_renderer(self.entry.get_used_size()))
        else :
            self.free_label.set_label(_("Not available"))
            self.avail_label.set_label(_("Not available"))
            self.used_label.set_label(_("Not available"))
        self.path_label.set_label(self.entry["FSTAB_PATH"])
        self.opt_label.set_label(escape_mkup(self.entry["FSTAB_OPTION"]))
        self.dev_label.set_label(self.entry["DEVICE"])
        self.label_label.set_label(self.disk.get_attribute(self.entry, "FS_LABEL"))
        self.uuid_label.set_label(self.disk.get_attribute(self.entry, "FS_UUID"))

        if self.disk.get_attribute(self.entry, "FS_VERSION") in "None":
            self.type_label.set_label("%s" % \
                    (self.disk.get_attribute(self.entry, "FS_TYPE")))
        else:
            self.type_label.set_label("%s (%s)" % \
                    (self.disk.get_attribute(self.entry, "FS_TYPE"), \
                    self.disk.get_attribute(self.entry, "FS_VERSION")))

        if self.disk.get_attribute(self.entry, "VENDOR") not in "None":
            self.vendor_label.set_label(self.disk.get_attribute(self.entry, "VENDOR"))

        if self.disk.get_attribute(self.entry, "MODEL") not in "None":
            self.model_label.set_label(self.disk.get_attribute(self.entry, "MODEL"))

        if self.disk.get_attribute(self.entry, "SERIAL") not in "None":
            self.serial_label.set_label(self.disk.get_attribute(self.entry, "SERIAL"))

        if self.disk.get_attribute(self.entry, "BUS") not in "None":
            self.bus_label.set_label(self.disk.get_attribute(self.entry, "BUS"))
        
    def on_quit(self, button, event = None) :
    
        self.dialog_info.hide()

class SanityCheck :

    def __init__(self, parent = None, conf = None) :
    
        self.parent = parent
        self.fstab = MntFile(FSTAB)
        self.conf = conf
        if not self.conf :
            self.conf = Config()
        
    def unknow(self) :

        unknow_block = []
        known = self.conf.get("Detected Device", "known").split()
        for entry in self.fstab.other :
            if re.search("(^UUID=|^LABEL=|^/dev/)", entry["FSTAB_NAME"]) \
                and entry["FSTAB_TYPE"] not in FstabData.ignore_fs \
                and entry["FSTAB_NAME"] not in FstabData.ignore_dev \
                and re.search("^/", entry["FSTAB_PATH"]) \
                and not entry["FSTAB_NAME"] in known :
                unknow_block.append(entry)
        if unknow_block :
            dial = dialog("warning", _("Removing unknown devices from\nthe configuration file?"), \
                [_("I can not find any existing block devices\n"
                "corresponding to the following devices:"), \
                _("It is advisable to remove them in order to avoid failed mount\n"
                "at start-up. Select the ones you want to remove.")], \
                [ "%s on %s" % (k["FSTAB_NAME"], k["FSTAB_PATH"]) for k in unknow_block ], \
                _("Remove selected"), [_("Don't show me this warning in the future.")], self.parent)
            if not dial[0] ==  Gtk.ResponseType.REJECT :
                for i in dial[1][0] :
                    logging.debug("Removing unknow entry : %s" % unknow_block[i]["FSTAB_NAME"])
                    self.fstab.other.remove(unknow_block[i])
                if dial[2][0] :
                    for i in dial[1][1] :
                        logging.debug("Always keep entry : %s" % unknow_block[i]["FSTAB_NAME"])
                        known.append(unknow_block[i]["FSTAB_NAME"])
                    self.conf.set("Detected Device", "known", " ".join(known))
                self.fstab.apply()      

    def duplicate(self) :

        checked = []
        for path in self.fstab.list(col = "FSTAB_PATH") :
            index = self.fstab.search(path)
            index.reverse()
            for i in index[:] :
                if self.fstab.count(self.fstab[i]) > 1 :
                    logging.warning("Removing exact duplicate entry : %s on %s" % \
                        (self.fstab[i]["DEVICE"], self.fstab[i]["FSTAB_PATH"]))
                    self.fstab.remove(i)
                    self.fstab.apply()
            index = self.fstab.search(path)
            index.reverse()
            if len(index) > 1 and not path in checked :
                entries = [ "%s on %s (%s)" % (self.fstab[k]["FSTAB_NAME"], \
                    self.fstab[k]["FSTAB_PATH"], self.fstab[k]["FSTAB_TYPE"]) for k in index ]
                data = []
                for i in range(len(index)) :
                    if self.fstab[index[i]]["FSTAB_PATH"] in device_is_mounted(self.fstab[index[i]]["DEVICE"]) :
                        data.append([entries[i], True])
                    else :
                        data.append([entries[i], False])
                dial = dialog("warning", _("Duplicate entry detected"), \
                [_("The following entries of your configuration file use the same mount point:"), \
                _("Two different devices can not share the same mount point.\n"
                "You should select the one you want, and the others will be commented.\n"
                "They will be kept in the configuration file, but will be disabled.")], \
                data, _("Keep selected"), parent = self.parent)
                if not dial[0] ==  Gtk.ResponseType.REJECT :
                    for j in dial[1][1] :
                        i = index[j]
                        logging.debug("Commenting duplicate : %s" % entries[j])
                        entry = ["#%s" % self.fstab[i]["FSTAB_NAME"]]
                        entry.extend([ self.fstab[i][u] for u in FstabData.categorie[1:] ])
                        self.fstab.remove(i)
                        self.fstab.comment.append(EntryBase(entry))
                    self.fstab.apply()
            checked.append(path)
        
        
def about_dialog(button, parent) :

    dial = Gtk.AboutDialog()
    dial.set_name(_("Disk Manager"))
    dial.set_version(VERSION)
    dial.set_copyright("Copyright © 2007 Mertens Florent \n 2021 MX Linux")
    dial.set_license( \
        "This program is free software; you can redistribute it and/or\n"
        "modify it under the terms of the GNU General Public License\n"
        "as published by the Free Software Foundation; either\n"
        "version 2 of the License, or (at your option) any later version.\n"
        "\n"
        "This program is distributed in the hope that it will be useful,\n"
        "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
        "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"
        "See the GNU General Public License for more details.\n"
        "\n"
        "You should have received a copy of the GNU General Public License\n"
        "along with this library; if not, write to the Free Software\n"
        "Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA")
    dial.set_comments(_("Easily manage your filesystem configuration"))
    dial.set_website(HOMEPAGE)
    dial.set_website_label("https://github.com/MX-Linux/disk-manager")
    dial.set_authors(AUTHORS)
    dial.set_translator_credits(_("translator-credits"))
    dial.set_logo_icon_name("disk-manager")
    dial.run()
    dial.hide()
          



# -*- coding: UTF-8 -*-
#
#  FstabData.py : Fstab data
#  Copyright (C) 2007 Mertens Florent <flomertens@gmail.com>
#  Updated 2021 for MX Linux Project by team member Nite Coder
#  Maintenance of project assumed by MX Linux with permission from original author.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
import os
import grp
from pwd import getpwnam
uid = getpwnam(os.getlogin())[2]
gid = grp.getgrnam('users')[2]

gvfs = ",x-gvfs-show"

# Default filesystem options
defaults = { "btrfs"        :   ("defaults" + gvfs, "0", "2"),
             "ext4"         :   ("defaults,noatime" + gvfs, "0", "2"),
             "ext3"         :   ("defaults,noatime" + gvfs, "0", "2"),
             "ext2"         :   ("defaults,noatime" + gvfs, "0", "2"),
             "exfat"        :   ("defaults" + ",uid=" + str(uid) + ",gid=" + str(gid) + ",dmask=0002,fmask=0113,utf8" + gvfs, "0", "2"),
             "exfat-fuse"   :   ("defaults" + ",uid=" + str(uid) + ",gid=" + str(gid) + ",dmask=0002,fmask=0113,iocharset=utf8,namecase=0,nonempty" + gvfs, "0", "2"),
             "vfat"         :   ("defaults" + ",uid=" + str(uid) + ",gid=" + str(gid) + ",dmask=0002,fmask=0113,utf8" + gvfs, "0", "2"),
             "ntfs"         :   ("defaults" + ",uid=" + str(uid) + ",gid=" + str(gid) + ",dmask=0002,fmask=0113,utf8" + gvfs, "0", "0"),
             "ntfs-3g"      :   ("defaults" + ",uid=" + str(uid) + ",gid=" + str(gid) + ",dmask=0002,fmask=0113,utf8" + gvfs, "0", "0"),
             "ntfs3"        :   ("defaults" + ",uid=" + str(uid) + ",gid=" + str(gid) + ",dmask=0002,fmask=0113,iocharset=utf8" + gvfs, "0", "0"),
             "jfs"          :   ("defaults,iocharset=utf8" + gvfs, "0", "0"),
             "__default__"  :   ("defaults" + gvfs, "0", "0")}

# Known special driver
special_driver = { "ntfs-3g"    : "Read-write driver",
                   "ntfs3"      : "Read-write kernel driver",
                   "ntfs"       : "Read-write driver",
                   "ntfs-fuse"  : "Read-write driver",
                   "__unknow__" : "Unknow driver" }

if os.path.exists('/sbin/mount.exfat-fuse'):
    special_driver["exfat-fuse"] = "exFAT-fuse driver"


# Secondary driver
secondary_driver = { "ext3"     : ("ext2"),
                     "exfat"    : ("exFAT"),
                     "vfat"     : ("msdos"),
                     "__all__"  : ("auto")}

# List type of device that have an FS_TYPE, but that we don't want to configure
ignore_fs = ("swap", "iso9660", "udf", "iso9660,udf", "udf,iso9660")

# List device that we should ignore
ignore_dev = ("/dev/fd0", "/dev/fd1", "/dev/fd2", "/dev/sr0", "/dev/sr1", "/dev/sr2")

# List of virtual device name
virtual_dev = ("proc", "devpts", "tmpfs", "sysfs", "shmfs", "usbfs")


# Common options. Keep them when we change fs
common = ("atime","noatime","diratime","nodiratime","auto","noauto","dev","nodev","exec",\
          "noexec","mand","nomand","user","nouser","users","group","_netdev","owner","suid","nosuid",\
          "ro","rw","sync","async","dirsync", "x-gvfs-show", "nofail")

# List of options that don't require a remount
dont_need_remount = ("auto", "noauto", "check=none", "nocheck", "errors=continue", "errors=remount-ro", "error=panic")


# Write entry in MntFile in this order :
path_order = ("/", "/usr", "/home")

# System partitions :
system = { "exact"  : ("/", "/home", "/tmp", "/boot", "/boot/efi", "/boot/grub"),
           "extand" : ("/usr", "/var", "/sys", "/proc")}


# MntFile header
header = "# /etc/fstab: static file system information.\n" + \
         "#\n" +\
         "# <file system> <mount point>   <type>  <options>       <dump>  <pass>\n\n"

# Categories of an entry
categorie = ("FSTAB_NAME", "FSTAB_PATH", "FSTAB_TYPE", "FSTAB_OPTION", "FSTAB_FREQ", "FSTAB_PASO")


# Divers
special_char = ('"', "'", " ", "<", ">", "&", "$", "(", ")", "`", "-", "|", ";", "~", "{", "}", "^")





# -*- coding: UTF-8 -*-
#
#  FstabError.py : Define Fstab exceptions
#  Copyright (C) 2007 Mertens Florent <flomertens@gmail.com>
#  Updated 2021 for MX Linux Project by team member Nite Coder
#  Maintenance of project assumed by MX Linux with permission from original author.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#


class NotInDatabase(Exception) : 

    def __init__(self, value) :
    
        self.value = value
        
    def __str__(self) :
        
        if isinstance(self.value, int) :
            return "Index %i out of range" % self.value
        else :
            return "Can't find '%s' in the database" % self.value
            
class DatabaseTypeError(Exception) :

    def __init__(self, type, accepted) :
    
        self.type = type
        self.accepted = accepted
        
    def __str__(self) :
    
        return "Wrong type : %s. Expected : %s" % \
            (self.type.__name__, ", ".join([ k.__name__ for k in self.accepted ]))
            
class UnknowEvent(Exception) :

    def __init__(self, event_handler, value) :

        self.value = value
        
    def __str__(self) :
    
        return "No event named %s" % self.value




# -*- coding: UTF-8 -*-
#
#  FstabDialogs.py : Provide dialogs for FstabHandler
#  Copyright (C) 2007 Mertens Florent <flomertens@gmail.com>
#  Updated 2021 for MX Linux Project by team member Nite Coder
#  Maintenance of project assumed by MX Linux with permission from original author.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from gi.repository import Gtk, Pango
#from gettext import gettext as _

from .Fstabconfig import *
from .SimpleGladeApp import SimpleGladeApp
import logging


ACCEPTED_TYPE = ["warning", "info", "error", "question"]


class WrongDialogType(Exception) : 

    def __init__(self, value) :
    
        self.value = value
        
    def __str__(self) :
        
        return "'%s' is not an accepted dialog type. " \
            "Should be '%s'" % (self.value, "' or '".join(ACCEPTED_TYPE))


class DialogBuilder(SimpleGladeApp) :
    ''' This class will create a builder with widget appropriate to the data you give hime '''

    def __init__(self, type, buttons, title, text, data, options, parent = None) :
        ''' See dialog doc for more information. You should not need to call this method
            directly, but use the dialog fct instead '''
    
        self.type = type
        self.buttons = buttons
        self.title = title
        self.text = text
        self.data = data
        self.options = options
        if not self.buttons :
            self.buttons = []
        elif isinstance(self.buttons, str) :
            self.buttons = [ self.buttons ]    
        if type in ("warning", "info", "error") :
            if not self.buttons :
                self.buttons = ["close"]
            else :
                self.buttons.insert(0, "cancel")
        elif type == "question" :
            self.buttons.insert(0, "yes")
            self.buttons.insert(0, "no")
        else :
            raise WrongDialogType(type)
        SimpleGladeApp.__init__(self, GLADEFILE, root = "dialog_template", domain = PACKAGE)
        self.dialog_main = self.main_widget
        self.dialog_main.set_title("")
        if parent :
            self.dialog_main.set_transient_for(parent)
        self.build_advance_dialog()
        self.set_size()
        
    def build_advance_dialog(self) :

        for button in self.buttons :
            if not isinstance(button, str) :
                continue
            try :
                response = getattr(Gtk.ResponseType, "%s" % button.upper())
            except :
                response = 0
            if hasattr(Gtk, "STOCK_%s" % button.upper()) :
                button = getattr(Gtk, "STOCK_%s" % button.upper())
            widget = self.dialog_main.add_button(button, response)
        self.dialog_main.set_focus(widget)
        self.title_label.set_label("<big><b>" + self.title + "</b></big>")
        try :
            self.image.set_from_stock(getattr(Gtk, "STOCK_DIALOG_%s" % self.type.upper()), Gtk.ICON_SIZE_DIALOG)
        except :
            pass
        if isinstance(self.text, list) :
            self.text_top_label.set_label(self.text[0])
            self.text_down_label.set_property("wrap", False)
            self.text_down_label.set_label(self.text[1])
        else :
            self.text_down_label.hide()
            self.text_top_label.set_label(self.text)
        if self.options :
            self.setup_check_button()
        if isinstance(self.data, list) :
            self.setup_treeview()
        elif isinstance(self.data, str) :
            self.setup_textview()
            
    def on_treeview_toggled(self, widget) :

        path = self.treeview.get_cursor()[0]  
        if path is None:
            return
        self.tree_store[path][0] = not self.tree_store[path][0]
    
    def setup_treeview(self) :
    
        self.treeview  = Gtk.TreeView()
        self.treeview.set_headers_visible(False)
        self.scrolledwindow.add(self.treeview)
        if len(self.data) > 1 and isinstance(self.data[1], bool) :
            sensitive = self.data[1]
            data = self.data[0]
        else :
            sensitive = True
            data = self.data
        renderer1 = Gtk.CellRendererToggle()
        renderer1.set_property("sensitive", sensitive)
        if sensitive :
            self.treeview.connect("cursor-changed", self.on_treeview_toggled)
        column = Gtk.TreeViewColumn("", renderer1, active=0)
        self.treeview.append_column(column)
        renderer2 = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn("", renderer2, text=1)
        self.treeview.append_column(column)
        self.tree_store = Gtk.ListStore(bool, str)
        self.treeview.set_model(self.tree_store)
        for entry in data :
            if isinstance(entry, list) :
                value = entry[1]
                entry = entry[0]
            else :
                value = True
            self.tree_store.append((value, entry))
        self.scrolledwindow.show_all()
        
    def setup_textview(self) :
    
        self.textview = Gtk.TextView()
        self.textview.set_editable(False)
        self.textview.set_cursor_visible(False)
        self.scrolledwindow.add(self.textview)
        buf = Gtk.TextBuffer()
        buf.set_text(self.data)
        self.textview.set_buffer(buf)
        self.scrolledwindow.show_all()
        
    def setup_check_button(self) :
    
        box = Gtk.VBox(homogeneous=True, spacing = 0)
        if len(self.options) > 1 and isinstance(self.options[1], bool) :
            grouped = self.options[1]
            options = self.options[0]
        else :
            grouped = False
            options = self.options
        widget = None
        for option in options :
            if isinstance(option, list) :
                option = option[0]
                value = option[1]
            else :
                value = False
            if grouped :
                widget = Gtk.RadioButton(label = option, group = widget)
            else :
                widget = Gtk.CheckButton(option)
            widget.set_active(value)
            #box.pack_start(widget, expand=False)
            box.pack_start(widget, False, False, 0)
        self.check_box = box
        #self.vbox.pack_end(self.check_box, expand=False)
        self.vbox.pack_end(self.check_box, False, False, 0)
        self.check_box.show_all()
        
    def set_size(self) :
        if hasattr(self, "treeview") :
            widget = self.treeview
            if self.data and len(self.data) > 1 and isinstance(self.data[1], bool) :
                sample = self.data[0]
            else :
                sample = self.data
            ypad = 6
            maxrow = 8
        elif hasattr(self, "textview") :
            widget = self.textview
            logging.debug('{}'.format(self.data))
            sample = self.data.split("\n")
            ypad = 2
            maxrow = 8
        else :
            return
        if sample is None:
            return
        
        ctx = widget.get_pango_context()
        layout = Pango.Layout(ctx)
        logging.debug('set_size: sample = {}'.format(sample))
        row = min(len(sample) + 2, maxrow)
        if self.options and row > 5 :
            row -= min(len(self.options), 3)
        w = []
        for k in sample :
            if isinstance(k, list) :
                k = k[0]
            layout.set_text(k + "O" * 2 * row)
            w.append(layout.get_pixel_size()[0])
        width = min(max(w), 450)
        height = (layout.get_pixel_size()[1] + ypad) * row
        widget.set_size_request(width, height)
    def run(self) :

        ret = self.dialog_main.run()
        if ret in (Gtk.ResponseType.DELETE_EVENT, Gtk.ResponseType.CANCEL, Gtk.ResponseType.CLOSE, Gtk.ResponseType.NO):
            ret = Gtk.ResponseType.REJECT
        result = [ret, None, None]
        if not ret == Gtk.ResponseType.REJECT :
            if hasattr(self, "treeview") and hasattr(self, "tree_store") :
                select = []
                unselect = []
                iter = self.tree_store.get_iter_first()
                for i in range(len(self.tree_store)) :
                    if self.tree_store.get_value(iter, 0) :
                        select.append(i)
                    else :
                        unselect.append(i)
                    iter = self.tree_store.iter_next(iter)
                result[1] = [select, unselect]
            if hasattr(self, "check_box") :
                select = []
                unselect = []
                i = 0
                for widget in self.check_box.get_children() :
                    if widget.get_active() :
                        select.append(i)
                    else :
                        unselect.append(i)
                    i += 1
                result[2] = [select, unselect]
        self.dialog_main.hide()
        return result

def dialog(type, title, text, data = None, action = None, options = None, parent = None) :
    ''' dialog(type, title, text, [data], [action], [options], [parent] -> launch a graphical dialog\n
        type : type of the dialog. should be "warning", "error", "info" or "question"
        title : title of the dialog
        text : the text to show in the dialog. If text = [text1, text2] text1 will be shown on top
               text2 will be shown on the bottom
        data : data to show in the dialog. If data is a list, show it in a treeview. otherwise,
               show it in a textview. 
               If treeview, list should be of the type : [[data, value], sensitive]
               sensitive : True/False, set the sensitivity of the check button. Default to True
               [data, value] : a list of data and it respectives values. Default value are False
               exemples : [ [data1, data2], False] : show a treeview with 2 unactivable, unselected row.
                          [data1, data2] : show a treeview with 2 activable, unselected row.
                          [[data1, True], [data2, False]] : show a treeview with 2 activable row, the first 
                                selected, the second unselected.
        options : list of options to show at the bottom of the dialog.
               options should be of the type : [[[option, value]], group]
               group : True/False, if True options are grouped and only one could be selected. Default to False
               [option, value] : a list of options and their respective values. Default values are False.
               See data for an exemple of how it works, it is nearly the same
        parent : the parent window\n
        return code : [dialog return code, [(selected data, unselected data)], [(selected opt, unselected opt)]]'''

    dialog = DialogBuilder(type, action, title, text, data, options, parent)
    return dialog.run()




# -*- coding: UTF-8 -*-
#
#  FstabHandler.py : High level class for managing fstab file
#  Copyright (C) 2007 Mertens Florent <flomertens@gmail.com>
#  Updated 2021 for MX Linux Project by team member Nite Coder
#  Maintenance of project assumed by MX Linux with permission from original author.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import os
import time
import shutil
import logging
import tempfile
import threading
import configparser
#from gettext import gettext as _

from .Fstabconfig import *
from .FstabUtility import *
from .Fstab import *
from .FstabDialogs import *
from .FstabError import *
from .EventHandler import *
from .Mounter import *
from . import FstabData

available_mode = ["real", "delay", "virtual"]

class FstabHandler(MntFile) :
    ''' FstabHandler(filename, [fd], [mode], [naming], [backend], [log], [parent]
                    [sanity_check]) -> new FstabHandler object\n
        FstabHandler is the high level class of MntFile. It try to provide high level methods.
        In particular, FstabHanler apply change automaticly, show a nice progress bar when
        mounting/unmouting device, can check for unknow and duplicate entry, integreate a logger,
        integrate an undo method... For now FstabHandler can only works with GTK, but
        in the future, Widget class will be abstracted to be able to use other backends\n
        Options are :
        filename, fd, mode, naming, backend : MntFile options. See Fstab.
            The only difference is that filaname is default to filename
        log : If True, save an original version of filename, and allow to use savelog method.
            Default to True.
        parent : specify the parent widget. Default to None
        sanity_check : if True, check for unknow and duplicate entry on creation.
            Default to False.
        external_change_watch : emit events on changes not due to this class
        mode : this is probably the more powerfull option of FstabHandler. 
            There is currently 3 mode available :
            real : the default mode. All changes are apply immediatly.
            delay : changes are only apply when calling the apply_now()
                methd, or when switching to the real mode.
            virtual : all changes are only apply when calling
                the apply_now() method. All not applied changes
                are lost when switching to an other mode. '''

    def __init__(self, filename = FSTAB, fd = None, mode = "real", naming = "auto", \
             backend = "auto", log = True, parent = None, external_change_watch = False) :

        self._parent = parent

        # Create a bckup of filename :
        bckup = "%s-%s-save" % (filename, PACKAGE)
        if not os.path.isfile(bckup) :
            shutil.copy(filename, bckup)

        # Load filename
        MntFile.__init__(self, filename, fd = fd, naming = naming, backend = backend)

        # Move it to fstab, and complete with other available device
        self._build_object()

        # Create copy
        self._current = self.copy()
        self._current.fstab = self.fstab.copy()
        self._original = self.fstab.copy()
        self._lastsave = self.fstab.copy()

        # Init log
        self.logchanges = log
        if not os.getuid() == 0 :
            self.logchanges = False
        self._initlog()

        # Set mode
        self.set_mode(mode)

        # Create Mounter object
        self._mounter = Mounter(self, parent = self._parent)
        self._mounter.hide_gui()

        # Start Event Handler
        self._event = EventHandler(self, external_change_watch)

    def _build_object(self, fstab = None) :

        logging.debug("Building FstabHandler...")
        t = time.time()
        if fstab :
            self[:] = fstab
        self.fstab = self.copy()
        self.fstab[:] = self[:]
        mtab = MntFile(MTAB, minimal = True)
        for device in self.info.list("DEVICE", ignored = False) :
            entries = [ mtab[k] for k in mtab.search(device, keys = ["DEVICE"]) ]
            if entries :
                for entry in entries :
                    path = entry["FSTAB_PATH"]
                    if path not in self.fstab.list("FSTAB_PATH") :
                        if entry["FSTAB_TYPE"] in ("fuse", "fuseblk") :
                            get_fuse_options(entry)
                        type = entry["FSTAB_TYPE"]
                        opt = entry["FSTAB_OPTION"].replace("rw", "defaults")
                        entry = self.add([device, path, type, opt])
                        logging.debug("-> Adding mounted device %s on %s" % (device, \
                                        entry["FSTAB_PATH"]))
            else :
                if device not in self.fstab.list() :
                    entry = self.add([device])
                    logging.debug("-> Adding %s on %s [type: %s]" % (device, entry["FSTAB_PATH"], entry["FSTAB_TYPE"]))
                    if entry["FSTAB_TYPE"] in "ntfs":
                       entry["FSTAB_TYPE"] = "ntfs-3g"
        logging.debug("FstabHandler build in %s s", time.time() -t)

    def _rebuild_object(self, fstab) :

        logging.debug("Rebuilding FstabHandler...")
        t = time.time()
        new = fstab.copy()
        new[:] = fstab[:]
        for entry in self[:] :
            if entry not in new :
                if not entry in self.fstab or not entry["DEVICE"] in new.list() \
                    or (entry["FSTAB_PATH"] not in new.list("FSTAB_PATH") and \
                            entry.get_is_mounted()) :
                    logging.debug("-> Adding %s on %s  (type: %s)" % (entry["DEVICE"], \
                            entry["FSTAB_PATH"], entry["FSTAB_TYPE"] ))
                    new.append(entry)
        self[:] = new
        self.fstab = fstab
        self._check_duplicate(self)
        logging.debug("FstabHandler rebuild in %s s", time.time() -t)

    def _copy(self) :

        new = self.fstab.copy()
        new.fstab = new.copy()
        new.fstab[:] = new[:]
        for entry in self :
            if entry not in new :
                new.append(entry.copy())
        new.current = self._current.copy()
        new.current.fstab = self._current.fstab.copy()
        return new

    def _restore_copy(self, copy) :

        self[:] = copy[:]
        self.fstab = copy.fstab
        self._current = copy.current
        self._current.fstab = copy.current.fstab

    def connect(self, event, fct, *kargs) :
        ''' x.connect(event, fct, *kargs) -> connect event to fct\n
            When event is detected, the function fct will be called with optional 
            parameters kargs. Available events are for now :
            - "external_fstab_changed" : emitted when filename is updated but not due to
                                         this class.
            - "external_mtab_changed" : emitted when MTAB is updated but not due to
                                        this class.
            - "external_changed" : emitted when one of the previous event is emitted
            - "internal_changed" : emitted when this class do a changed 
                                   (mount/unmount/configure...)
            - "configuration_changed" : emitted when one of the previous event is emitted
            - "size_changed" : emitted when size of all mounted device changed significantly
            - "any_changed" : emitted when one of the previous event is emitted '''

        if hasattr(self._event, "on_%s" % event) :
            logging.debug("Connecting event '%s' to %s%s" % (event, fct.__name__, kargs))
            setattr(self._event, "%s_fct" % event, [fct, kargs])
            setattr(self, "%s_fct" % event, [fct, kargs])
        else :
            raise UnknowEvent(self._event, event)

    def shutdown(self) :
        ''' x.shutdown() -> Shutdown FstabHandler process and save log\n
            It is important to call this function if you use the external change watch
            to shutdown it gracefully.'''

        logging.debug("Calling Shutdown...")
        if self.lastsave_has_changed() :
            self.savelog()
        list_created_path("clean")
        self._event.shutdown()

    def set_mode(self, mode) :
        ''' x.set_mode(mode) -> set mode to mode\n
            Be aware that when quiting a virtual mode, all Entry of x are recreated '''

        if not mode in available_mode :
            logging.warning("No FstabHanler mode named %s. Starting with 'real' mode" % mode)
            self._mode = "real"
        if hasattr(self, "_mode") :
            self._previous_mode = self._mode
            t = time.time()
            logging.debug("Switching mode from %s to %s" % (self._mode, mode))
            if self._mode == "virtual" and not mode == "virtual" and hasattr(self, "_real") :
                logging.debug("Restoring state...")
                self._restore_copy(self._real)
            elif self._mode == "delay" and not mode == "delay" :
                logging.debug("Exiting from delay mode. Applying changes...")
                self.apply_now()
            if not self._mode == "virtual" and mode == "virtual" : 
                self._real = self._copy()
            logging.debug("Mode switched in %s s", time.time() -t)
        else :
            logging.debug("Starting in %s mode" % mode)
        self._mode = mode
        
    def set_previous_mode(self) :
        ''' x.set_previous_mode() -> set mode to it's previous value '''
    
        if hasattr(self, "_previous_mode") :
            self.set_mode(self._previous_mode)

    def get_mode(self) :
        ''' x.get_mode() -> get the current mode '''
    
        return self._mode
            
    def get_configured(self) :
        ''' x.get_configured() -> get all device configured in filename '''
        
        return [ k for k in self.fstab ]
        
    def get_all(self) :
        ''' x.get_all() -> get all manageable devices '''
    
        return [ k for k in self ]
        
    def get_new(self) :
        ''' x.get_new() -> get all manageable devices non configures in filename '''
    
        return [ k for k in self if k not in self.fstab ]
        
    def get_duplicate(self, item) :
        ''' x.get_duplicate(item) -> get duplicate entry of item\n
            item might be an entry or a device name.
            if in duplicate, return a list of entry with the same dev, including iteself.
            return ann empty list otherwise '''
            
        if isinstance(item, Entry) :
            dev = item["DEVICE"]
        else :
            dev = item
        res = self.search(dev, keys = ["DEV", "DEVICE"])
        if len(res) > 1 :
            return [ self[k] for k in res ]
        else :
            return []
        
    def _get_entry(self, item) :
    
        if isinstance(item, Entry) :
            return item
        else :
            return self[item]
        
    def get_attribute(self, item, attribute) :
        ''' x.get_attribute(item, attribute) -> get the attribute of item\n
            item can be a FSTAB_PATH or an entry '''
    
        entry = self._get_entry(item)
        if attribute in entry :
            return entry[attribute]
        return "None"
        
    def get_property(self, item, property) :
        ''' x.get_property(item, attribute) -> get a property of item\n
            item can be a FSTAB_PATH or an entry
            current available property are :
            - is_system
            - is_mounted
            - size, free_size, availale_size, used_size '''
    
        entry = self._get_entry(item)
        if hasattr(entry, "get_" + property) :
            return getattr(entry, "get_" + property)()
        return None

    def set(self, item, path=None, type=None, option=None, paso=None) :
        ''' x.set(item, [path], [type], [option], [paso]) -> set item path/type/option 
                                                             and/or paso\n
            item can be a FSTAB_PATH or an entry '''

        entry = self._get_entry(item)
        if path :
            old_path = entry["FSTAB_PATH"]
            entry["FSTAB_PATH"] = path
        if type :
            entry["FSTAB_TYPE"] = type
        if option :
            if option[0] == "+" :
                entry.addopt(option[1:])
            elif option[0] == "-" :
                entry.removeopt(option[1:])
            else :
                entry.setopt(option)
        if not paso == None :
            entry["FSTAB_PASO"] = str(int(bool(paso)) + \
                int(bool(paso) and not entry.get_is_system()))
        self._apply()
            
    def set_default_option(self, item) :
        ''' x.set_default_option(item) -> set item options to default\n
            item can be a a FSTAB_PATH or an entry '''
    
        entry = self._get_entry(item)
        entry["FSTAB_OPTION"] = entry.defaultopt()
          
    def mount(self, item) :
        ''' x.mount(item) -> mount item\n
            item can be a FSTAB_PATH or an entry '''
    
        entry = self._get_entry(item)
        if entry.get_is_mounted() :
            logging.debug("%s on %s already mounted" % (entry["DEVICE"], \
                            entry["FSTAB_PATH"]))
            return 0
        self._event.emit("internal_changed_prepare")
        result = self._mounter.do(mount = [entry])
        if not result[1] :
            if not self._current == self or not self._current.fstab == self.fstab :
                logging.debug("Detected changement.")
                self.simple_apply()
        else :
            clean_path(entry["FSTAB_PATH"])
        self._event.emit("internal_changed")
        return max(result)

    def umount(self, item) :
        ''' x.umount(item) -> unmount item\n
            item can be a FSTAB_PATH or an entry '''
    
        entry = self._get_entry(item)
        if not entry.get_is_mounted() :
            logging.debug("%s on %s already unmounted" % (entry["DEVICE"], \
                            entry["FSTAB_PATH"]))
            return 0
        self._event.emit("internal_changed_prepare")
        result = self._mounter.do(umount = [entry])
        if not result[0] :
            self._check_duplicate([entry])
            clean_path(entry["FSTAB_PATH"])
            self._event.emit("internal_changed")
        else :
            self._event.emit("internal_changed_cancel")
        return max(result)
        
    def configure(self, item) :
        ''' x.configure(item) -> add item to filename and mount it\n
            item can be a FSTAB_PATH or an entry '''
    
        entry = self._get_entry(item)
        if not entry in self.fstab :
            self.fstab.append(entry)
            self._apply()
        
    def unconfigure(self, item) :
        ''' x.unconfigure(item) -> remove item from filename and ask user if 
                                   he want to unmount it\n
            item can be a FSTAB_PATH or an entry '''
    
        entry = self._get_entry(item)
        if entry in self.fstab :
            self.fstab.remove(entry)
            self._apply()
            self._check_duplicate([entry])
                
    def _check_duplicate(self, entries) :
    
        for entry in entries[:] :
            if not entry.get_is_mounted() \
                    and len(self.search(entry["DEV"], keys = ["DEV"])) > 1 \
                    and not entry in self.fstab and entry in self :
                self.remove(entry)
                logging.debug("Removing unused duplicate : %s on %s" \
                        % (entry["DEVICE"], entry["FSTAB_PATH"]))
                if entry in self._current :
                    self._current.remove(entry)

    def _apply(self) :

        # Exit if we are in virtual or delay mode
        if self._mode == "virtual" or self._mode == "delay" :
            logging.debug("Mode %s enable, can't apply now" % self._mode)
            logging.debug("Expected changes :\n%s" % self._get_state())
            return
            
        # inhibit the watch so we don't trigger it while we apply the change
        self._event.emit("internal_changed_prepare")
        logging.debug("Applying changes")
        
        # Detect changes via an ugly hack that track global & local change
        change = change_tracker(self._current, self)
        change2 = change_tracker(self._current.fstab, self.fstab)
        for i in range(len(change)) :
            [ change[i].append(k) for k in change2[i] if k not in change[i] ]
        
        # Apply the change now
        self.simple_apply()
        
        # Create path if needed
        self.fstab.make_all_path()
        
        # Mount entry added and changed
        to_mount = change[1]
        logging.debug("-> Device to mount :\n-> |%s" % \
            "\n-> |".join([ "%s on %s" % (k["DEVICE"], k["FSTAB_PATH"]) for k in to_mount ]))
                
        # Unmount entry changed, and ask user for non duplicate disabled entry
        to_umount = change[2]
        disabled = change[3][:]
        if disabled :
            dial = dialog("warning", _("Unmounting disabled devices?"), \
                [_("You disabled the following devices:"), _("Do you want to unmount them?")], \
                [ "%s on %s" % (k["DEV"], k["FSTAB_PATH"]) for k in change[3] ], \
                _("Unmount selected"), parent = self._parent)
            if not dial[0] == Gtk.ResponseType.REJECT :
                for i in dial[1][0] :
                    to_umount.append(change[3][i])
        logging.debug("-> Device to umount :\n-> |%s" % \
            "\n-> |".join([ "%s on %s" % (k["DEVICE"], k["FSTAB_PATH"]) for k in to_umount ]))

        # Require reboot if some device couldn't be unmounted
        if self._mounter.do(mount = to_mount, umount = to_umount)[0] :
            dialog("warning", _("Reboot required"), \
                _("In order to apply all your changes,\n"\
                "you'll need to reboot your computer."), parent = self._parent)
                
        # Remove duplicate that are no more needed
        self._check_duplicate(to_umount)
                    
        # Delete uneeded path created here
        clean_all_path()
                
        # If state changed between the apply and here, do a simple_apply.
        # This can be due to a FSTAB_PATH change during the mounting
        if not self._current == self or not self._current.fstab == self.fstab :
            logging.debug("Detected changement.")
            self.simple_apply()
            
        # Emit internal_changed event
        self._event.emit("internal_changed")
                    
    def apply_now(self) :
        ''' x.apply_now() -> apply the changes now\n
            should only be used in virtual or delay mode, since
            all changes are automaticly apply in real mode '''
    
        if self._current == self and self._current.fstab == self.fstab :
            logging.debug("Nothing to apply")
            return
        mode = self._mode
        self._mode = "real"
        self._apply()
        self._mode = mode
        if self._mode == "virtual" :
            self._real = self._copy()
            
    def simple_apply(self) :
        ''' x.simple_apply() -> apply change without mounting/unmounting/emmtting event \n '''
        
        logging.debug("Updating current situation:\n%s" % self._get_state())
        # Apply the change now
        MntFile.apply(self.fstab)
        
        # Update current situtation
        self._current = self.copy()
        self._current.fstab = self.fstab.copy()
        
    def _get_state(self, obj = None) :
    
        if not obj :
            obj = self
        change1 = change_tracker(obj._current, obj)[0]
        change2 = change_tracker(obj._current.fstab, obj.fstab)[0]
        return "-> Global change :\n-> |%s\n-> Local change :\n-> |%s" % \
                ("\n-> |".join([ k.replace("\n","") for k in change1 ]), \
                "\n-> |".join([ k.replace("\n","") for k in change2 ]))

    def _initlog(self) :
    
        if self.logchanges :
            self._logconf = configparser.RawConfigParser()
            self._logconf.read(FSTAB_LOG)
            if not self._logconf.has_section("Original") :
                logging.debug("Creating original backup")
                self._logconf.add_section("Original")
                self._logconf.set("Original","log","Original configuration")
                fstabfile = open(self.filename)
                self._logconf.set("Original","fstab",fstabfile.read())
                fstabfile.close()
                self._logfile = open(FSTAB_LOG, "w")
                self._logconf.write(self._logfile)
                self._logfile.close()        
                     
    def savelog(self, name = None) :
        ''' x.savelog(name) -> save log as name\n
            name is default to the date of saving '''
    
        if not name :
            name = str(time.time())
        if self.logchanges and self.lastsave_has_changed() :
            logging.debug("Save log as %s", name)
            self._logconf.add_section(name)
            log = change_tracker(self._lastsave, self.fstab)[0]
            self._logconf.set(name, "log", "\n".join(log).strip())
            self._logconf.set(name, "fstab", self.fstab.write())
            self._logfile = open(FSTAB_LOG, "w+")
            self._logconf.write(self._logfile)
            self._logfile.close()
            self._lastsave = self.fstab.copy()
            self._initlog()
            
    def get_listlog(self) :
        ''' x.listlog() -> list all the current log recorded '''
    
        return self._logconf.sections()
        
    def get_logcommit(self, name) :
        ''' x.get_commit(name) -> get commit of log name '''
        
        return self._logconf.get(name, "log")
        
    def get_logfile(self, name) :
        ''' x.get_logfile(name) -> get file of log name '''
        
        return self._logconf.get(name, "fstab")
        
    def get_changes_current_from(self, name) :
        ''' x.get_changes_current_from(name) -> get log of change between name version -> current version '''
    
        tmpfile = tempfile.NamedTemporaryFile('w+t')
        tmpfile.write(self._logconf.get(name, "fstab"))
        tmpfile.seek(0)
        previous = MntFile(self.filename, tmpfile)
        return change_tracker(previous, self.fstab)[0]
        
    def get_changes_current_to(self, name) :
        ''' x.get_changes_current_to(name) -> get log of change between current version -> name version '''
    
        tmpfile = tempfile.NamedTemporaryFile('w+t')
        tmpfile.write(self._logconf.get(name, "fstab"))
        tmpfile.seek(0)
        previous = MntFile(self.filename, tmpfile)
        return change_tracker(self.fstab, previous)[0]
            
    def revert_to(self, name) :
        ''' x.revert_to(name) -> revert to version name '''
    
        logging.debug("Revert version to %s", name)
        tmpfile = tempfile.NamedTemporaryFile('w+t')
        tmpfile.write(self._logconf.get(name, "fstab"))
        tmpfile.seek(0)
        self._rebuild_object(MntFile(self.filename, tmpfile))
        self._apply()
        
    def cleanlog(self) :
        ''' x.cleanlog() -> clean the log and save current filename as Original '''
    
        self._logfile = open(FSTAB_LOG, "w")
        self._logfile.close()
        self._initlog()

    def get_changes_current_to_original(self) :
    
        return change_tracker(self.fstab, self._original)[0]
        
    def undo(self) :
        ''' x.undo() -> undo all changes made since the creation of the object '''

        self._rebuild_object(self._original.copy())
        self._apply()
        
    def original_has_changed(self) :
        ''' x.original_has_changed() -> return True if the original version at
                                        the creation of the object has changed '''
    
        return bool(not self.fstab == self._original)
        
    def lastsave_has_changed(self) :
        ''' x.original_has_changed() -> return True if the version has changed since
                                        last savelog '''

        return bool(not self.fstab == self._lastsave)




# -*- coding: UTF-8 -*-
#
#  Utility.py : Various utility
#  Copyright (C) 2007 Mertens Florent <flomertens@gmail.com>
#  Updated 2021 for MX Linux Project by team member Nite Coder
#  Maintenance of project assumed by MX Linux with permission from original author.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import os
import pwd
import logging
from subprocess import *
from urllib.parse import quote

#from gettext import gettext as _

from .config import *

pref_browser = { "GNOME" : { "file" : [["gnome-open"], ["nautilus", "--no-desktop"]], \
                             "http" : [["gnome-open"], ["epiphany"], ["firefox"], ["mozilla"]] }, \
                 "KDE"   : { "file" : [["kfmclient", "exec"], ["konqueror"], ["dolphin"]], \
                             "http" : [["kfmclient", "exec"], ["konqueror"], ["opera"]] }, \
                 "XFCE"  : { "file" : [["exo-open"], ["thunar"], ["nautilus", "--no-desktop"]], \
                             "http" : [["exo-open"], ["firefox"], ["mozilla"], ["epiphany"]]}}

def size_renderer(size) :
    ''' Return a string that show the size in a human readable way '''

    if not size : 
        return "0"
    for unit in [_("KB"), _("MB"), _("GB"), _("TB") ] :
        size = float(size)/1024.
        if size/1024. < 1 :
            return str(round(size, 1)) + " " + unit
    return str(round(size, 1)) + " " + unit

def get_user(request = "name") :
    ''' Get user information. Request could be name, uid, gid, dir, shell, ... '''

    user = pwd.getpwnam(os.getlogin())

    if user.pw_uid == "0" :
        logging.warning("Can't find your real username. Some features will not work.")
    return getattr(user, "pw_" + request)

def open_url(url) :
    ''' Select a browser depending on the requested url, and on the desktop currently used.
        We also do some ugly tweaks to be able to execute this browser as non root '''

    if url[:7] == "http://" :
        url = "http://%s" % quote(url.split("http://")[1])
    elif url[:7] == "file://" :
        url = "file://%s" % quote(url.split("file://")[1])
    else :
        url = "file://%s" % quote(url)
    user = get_user()
    desktop = get_desktop(default = "GNOME")
    browsers= pref_browser[desktop][url[:4]]
    [ browsers.extend(k) for k in [ k[url[:4]] for k in list(pref_browser.values()) ] ]
    browser = ""
    for test in browsers :
        if test_cmd(test[0]) :
            browser = " ".join(test)
            break
    if not browser:
        browser = "xdg-open"
        
    logging.debug("Launching %s as user %s with %s" % (url, user, browser))

    if not browser or not user or user == "root" :
        logging.debug("-> Not possible")
        return False
    for key in ["HOME", "USER", "XAUTHORITY", "LOGNAME", "MAIL", "XDG_RUNTIME_DIR"] :
        if key in os.environ :
            setattr(open_url, "current_%s" % key, os.environ[key])
    os.putenv("USER", user)
    os.putenv("LOGNAME", user)
    os.putenv("HOME", get_user("dir"))
    os.putenv("MAIL", "/var/mail/%s" % user)
    os.putenv("XDG_RUNTIME_DIR", "/run/user/%s" % str(get_user("uid")))
    if "%s_user-XAUTHORITY" % PACKAGE in os.environ :
        os.putenv("XAUTHORITY", os.environ["%s_user-XAUTHORITY" % PACKAGE])

    cmd = "runuser -u %s %s  '%s'" % (user, browser, url)

    logging.debug("Command : %s" % cmd)
    #run(cmd, capture_output=False)
    Popen(cmd, close_fds=True, shell=True)

    for key in ["HOME", "USER", "XAUTHORITY", "LOGNAME", "MAIL", "XDG_RUNTIME_DIR"] :
        if hasattr(open_url, "current_%s" % key) :
            os.putenv(key, getattr(open_url, "current_%s" % key))
    return True

def test_cmd(cmd) :

    return not call(["which " + cmd], stdout=PIPE, stderr=PIPE, \
                        close_fds=True, shell=True)
    
def pidof(proc_name) :

    return not call(["pidof " + proc_name], stdout=PIPE, stderr=PIPE, \
                        close_fds=True, shell=True)

def get_desktop(default = None) :
    ''' Get current running desktop. Search first for an environment key, otherwise for 
        a process. This way we are not fool by people running a gnome-panel in a kde 
        session for example, but we still can detect a desktop if keys are unset by a su
        helper (conseolheper for example). '''

    if "KDE_FULL_SESSION" in os.environ or "KDE_MULTIHEAD" in os.environ:
        desktop = "KDE"
    elif "GNOME_DESKTOP_SESSION_ID" in os.environ :
        desktop = "GNOME"
    elif pidof("startkde") or pidof("kicker") :
        desktop = "KDE"
    elif pidof("gnome-session") or pidof("gnome-panel") :
        desktop = "GNOME"
    elif pidof("xfce4-session") or pidof("xfce-mcs-manager") :
        desktop = "XFCE"
    else :
        logging.debug("No desktop environement found. Using default one : %s" % default)
        desktop = default
    logging.debug("Detected %s session" % desktop)
    return desktop   




# -*- coding: UTF-8 -*-
#
#  DiskManager.py : Manage main interfaces
#  Copyright (C) 2007 Mertens Florent <flomertens@gmail.com>
#  Updated 2021 for MX Linux Project by team member Nite Coder
#  Maintenance of project assumed by MX Linux with permission from original author.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import os
import re
import sys
import time
import logging
import threading
from subprocess import Popen
from xml.sax.saxutils import escape as escape_mkup

import gi
gi.require_version("Gtk", "3.0")
gi.require_version("GdkPixbuf", "2.0")
from gi.repository import Gtk
from gi.repository import GdkPixbuf, Pango
#  from .SimpleGladeApp import SimpleGladeApp
#from gettext import gettext as _
#import gettext
#gettext.install('disk-manager', '/usr/share/locale')

from DiskManager.Utility import *

TRAYTYPE = "gtk"
NOTIFICATION = False

ICON_THEME = Gtk.IconTheme.get_default()
ICON_THEME.append_search_path("%s/.icons" % get_user("dir"))
ICON_THEME.append_search_path("%s/.local/share/icons" % get_user("dir"))
try :
    PIX_MOUNT = ICON_THEME.load_icon("gnome-dev-harddisk", 24, 0)
except :
    try :
        PIX_MOUNT = ICON_THEME.load_icon("hdd_mount", 24, 0)
    except :
        PIX_MOUNT = GdkPixbuf.Pixbuf(GdkPixbuf.Colorspace.RGB, True, 8, 1, 1)
try :
    PIX_UNMOUNT = ICON_THEME.load_icon("gnome-dev-removable", 24, 0)
except :
    try :
        PIX_UNMOUNT = ICON_THEME.load_icon("hdd_unmount", 24, 0)
    except :
        PIX_UNMOUNT =  GdkPixbuf.Pixbuf(GdkPixbuf.Colorspace.RGB, True, 8, 1, 1)

from DiskManager.config import *
from DiskManager.Config import Config
from DiskManager.Dialogs import *
from DiskManager.Fstab.FstabHandler import *


class DiskManager(SimpleGladeApp):
    ''' Class that manage the main window '''

    def __init__(self) :

        SimpleGladeApp.__init__(self, GLADEFILE, "window_main", domain = "disk-manager")

        self.window_main.set_title(_("Disk Manager"))

        # Start configuration
        self.conf = Config()

        # Perform sanity check
        check = SanityCheck(parent = self.window_main, conf = self.conf)
        check.unknow()
        check.duplicate()

        # Start fstab handler
        naming = self.conf.get("General", "fstab_naming")
        backend = self.conf.get("General", "backend")
        self.disk = FstabHandler(FSTAB, naming = naming, backend = backend,
                parent = self.window_main, external_change_watch = True)
        # Draw stuff
        try :
            self.window_main.resize(
                self.conf.getint("Gui Config", "main_width"),
                self.conf.getint("Gui Config", "main_height"))
        except :
            self.conf.set_default("Gui Config", "main_width")
            self.conf.set_default("Gui Config", "main_height")
        self.width = self.conf.getint("Gui Config", "main_width")
        self.height = self.conf.getint("Gui Config", "main_width")
        self.window_main.show()
        self.setupTreeView()
        self.update_main()
        # Store all detected device
        detected = " ".join([ k["DEVICE"] for k in self.disk.get_all() ])
        self.conf.set("Detected Device", "detected", detected)

        # Connect events
        self.gtkbuttonquit.connect("clicked", self.on_close_clicked)
        self.window_main.connect("destroy", self.on_close_clicked)
        self.window_main.connect("configure-event", self.on_resize)
        #self.undo_menu.connect("activate", self.on_revert_clicked)
        #self.save_menu.connect("activate", self.on_save_clicked)
        #self.history_menu.connect("activate", self.on_history_clicked)
        self.gtkbuttonabout.connect("clicked", about_dialog, self.window_main)
        self.edit_button.connect("clicked",  self.on_edit_clicked)
        self.mount_button.connect("clicked",  self.on_mount_clicked)
        self.info_button.connect("clicked", self.on_info_clicked)
        self.treeview2.connect("row-activated", self.on_row_activated)
        self.treeview2.connect("cursor_changed", self.on_cursor_changed)
        self.disk.connect("any_changed", GObject.idle_add, self.update_main)
        self.gtkbuttonhelp.connect("clicked", self.on_help_clicked)

    def setupTreeView(self) :
        ''' treeview setup '''

        renderer= Gtk.CellRendererToggle()
        renderer.set_property("activatable", True)
        renderer.connect("toggled", self.on_enable_toggled)
        column = Gtk.TreeViewColumn(_("Enable"), renderer, active=0, sensitive=1, activatable=1)
        column.set_sort_column_id(0)
        column.set_reorderable(True)
        self.treeview2.append_column(column)

        pix_renderer = Gtk.CellRendererPixbuf()
        column = Gtk.TreeViewColumn(_("Device"))
        column.pack_start(pix_renderer, False)
        column.set_attributes(pix_renderer, pixbuf=6)
        pix_renderer.set_property("xpad", 3)
        text_renderer = Gtk.CellRendererText()
        column.pack_end(text_renderer, False)
        column.set_attributes(text_renderer, text=2, sensitive=1, style=5)
        column.set_sort_column_id(2)
        column.set_reorderable(True)
        column.set_expand(True)
        self.treeview2.append_column(column)
        
        renderer = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn(_("Mount point"), renderer, text=3, sensitive=1, style=5)
        column.set_sort_column_id(3)
        column.set_expand(True)
        column.set_reorderable(True)
        self.treeview2.append_column(column)
       
        renderer = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn(_("Type"), renderer, text=4, sensitive=1, style=5)
        column.set_sort_column_id(4)
        column.set_expand(True)
        column.set_reorderable(True)
        self.treeview2.append_column(column)
        
        renderer = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn(_("Total"), renderer, text=7, sensitive=1, style=5)
        column.set_sort_column_id(7)
        column.set_expand(True)
        column.set_reorderable(True)
        self.treeview2.append_column(column)

        renderer = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn(_("Free"), renderer, text=8, sensitive=1, style=5)
        column.set_sort_column_id(8)
        column.set_expand(True)
        column.set_reorderable(True)
        self.treeview2.append_column(column)

        renderer = Gtk.CellRendererProgress()
        column = Gtk.TreeViewColumn(_("Used"), renderer, text=9, value=10, sensitive=1)
        column.set_sort_column_id(9)
        column.set_expand(True)
        column.set_reorderable(True)
        renderer.set_property("ypad", 4)
        renderer.set_property("xpad", 4)
        self.treeview2.append_column(column)

        self.tree_store = Gtk.ListStore(bool, bool, str, str, str, int, \
                    GdkPixbuf.Pixbuf, str, str, str, int)
        self.tree_store.set_sort_func(7, self.sort_size, "size")
        self.tree_store.set_sort_func(8, self.sort_size, "free")
        self.tree_store.set_sort_func(9, self.sort_size, "used")
        try :
            self.tree_store.set_sort_column_id( \
                int(self.conf.get("Gui Config", "selected")), \
                getattr(Gtk.SortType, self.conf.get("Gui Config", \
                    "selected_order").upper()))
        except ValueError :
            self.tree_store.set_sort_column_id( \
                int(self.conf.set_default("Gui Config", "selected")), \
                getattr(Gtk.SortType, self.conf.set_default("Gui Config", \
                    "selected_order").upper()))
        self.treeview2.set_model(self.tree_store)

    def update_main(self) :
        ''' Update the window '''

        focus_path = self.treeview2.get_cursor()[0]
        if focus_path :
            l = self.tree_store[focus_path]
            current_focus = (l[2], l[3])

        #self.undo_menu.set_sensitive(self.disk.original_has_changed())
        #self.save_menu.set_sensitive(self.disk.lastsave_has_changed())

        (self.total, self.total_free, self.total_used, self.total_avail) = (0,0,0,0)
        configured = self.disk.get_configured()
        self.tree_store.clear()
        for entry in self.disk :
            device = entry["DEV"]
            type = entry["FSTAB_TYPE"]
            path = entry["FSTAB_PATH"]
            sensitive = not entry.get_is_system() or not entry in configured
            size_real  = entry.get_size()
            size = size_renderer(size_real)
            if entry in configured :
                self.total += size_real
                enable = True
                style = Pango.Style.NORMAL
            else :
                enable = False
                style = Pango.Style.ITALIC
            if entry.get_is_mounted() :
                pix = PIX_MOUNT
                free_real = entry.get_free_size()
                free = size_renderer(free_real)
                used_real = entry.get_used_size()
                used = size_renderer(used_real)
                try :
                    per_use = used_real  / float(size_real) * 100
                except ZeroDivisionError :
                    per_use = 0
                if entry in configured :
                    self.total_free += free_real 
                    self.total_used += used_real 
                    self.total_avail += entry.get_available_size()
            else :
                pix = PIX_UNMOUNT
                free = used = ""
                per_use = 0
            self.tree_store.append((enable, sensitive, device, path, type, \
                    style, pix, size, free, used, per_use))

        if focus_path :
            for path in range(len(self.tree_store)) :
                l = self.tree_store[path]
                if current_focus == (l[2], l[3]) :
                    self.treeview2.set_cursor(path)
                    break
                elif current_focus[0] == l[2] and len(self.disk.search(l[2], keys = ["DEV"])) < 2 :
                    self.treeview2.set_cursor(path)
                    break
        if not self.treeview2.get_cursor()[0] :
            self.info_button.set_sensitive(False)
            self.mount_button.set_sensitive(False)
            self.edit_button.set_sensitive(False)

    def sort_size(self, treemodel, iter1, iter2, type) :
    
        iter = [iter1, iter2]
        size = [0, 0]
        for i in range(2) :
            if not self.tree_store.get_value(iter[i], 3) :
                return 0
            entry = self.tree_store.get_value(iter[i], 3)
            if type == "size" :
                size[i] = self.disk.get_property(entry, type)
            else :
                size[i] = self.disk.get_property(entry, "%s_size" % type)
        if size[0] < size[1] :
            return -1
        return 1
             
    def on_edit_clicked(self, button) :

        path = self.treeview2.get_cursor()[0]
        entry = self.disk[self.tree_store[path][3]]
        if entry.get_is_system() :
            question = Gtk.MessageDialog(
                flags=0,
                message_type=Gtk.MessageType.QUESTION,
                buttons=Gtk.ButtonsType.YES_NO,
                text=_("Editing system partition?"),
                ) 
            question.format_secondary_text(_("%s is an important system partition.\n" \
                "Be really careful when editing it, or you may have\n" \
                "serious problems. Do you want to continue?")%entry["DEV"])
            response = question.run()
            question.destroy()      
            if response == Gtk.ResponseType.NO :
                return
                
        dial = EditPartition(self.disk, entry, parent = self.window_main)
        dial.dialog_edit.run()

    def on_mount_clicked(self, button) :
    
        path = self.treeview2.get_cursor()[0]
        entry = self.disk[self.tree_store[path][3]]
        if entry.get_is_mounted() : 
            self.disk.umount(entry)
        else : 
            self.disk.mount(entry)
            
    def on_help_clicked(self, button) :
        
        help_url = "https://mxlinux.org/wiki/help-files/help-disk-manager/"

        user = get_user()
        
        cmd = "xdg-open %s" % help_url
        if os.path.exists("/usr/bin/xlinks2") :
           cmd = "/usr/bin/xlinks2 %s" % help_url
        if os.path.exists("/usr/bin/mx-viewer") :
           cmd = "/usr/bin/mx-viewer %s disk-manager" % help_url
        if os.path.exists("/usr/bin/antix-viewer") :
           cmd = "/usr/bin/antix-viewer %s disk-manager" % help_url

        cmd = "runuser -u %s  -- " % user + cmd

        for key in ["HOME", "USER", "XAUTHORITY", "LOGNAME", "MAIL", "XDG_RUNTIME_DIR"] :
            if key in os.environ :
                setattr(open_url, "current_%s" % key, os.environ[key])
        os.putenv("USER", user)
        os.putenv("LOGNAME", user)
        os.putenv("HOME", get_user("dir"))
        os.putenv("MAIL", "/var/mail/%s" % user)
        os.putenv("XDG_RUNTIME_DIR", "/run/user/%s" % str(get_user("uid")))
        if "%s_user-XAUTHORITY" % PACKAGE in os.environ :
            os.putenv("XAUTHORITY", os.environ["%s_user-XAUTHORITY" % PACKAGE])
    
        logging.debug("Command : %s" % cmd)
        Popen(cmd, close_fds=True, shell=True)
    
        for key in ["HOME", "USER", "XAUTHORITY", "LOGNAME", "MAIL", "XDG_RUNTIME_DIR"] :
            if hasattr(open_url, "current_%s" % key) :
                os.putenv(key, getattr(open_url, "current_%s" % key))
        return True
        
    def on_info_clicked(self, button) :
    
        path = self.treeview2.get_cursor()[0]
        col = path[0]
        dial = InfoDialog(None, self.disk, parent = self.window_main)
        dial.entry = self.disk[self.tree_store[path][3]]
        dial.update_dial()
        res = dial.dialog_info.run()
        path = (col,)
        self.treeview2.set_cursor(path)
        dial.dialog_info.destroy()
        
    def on_row_activated(self, treeview, path, view_column) :
        ''' Start browser on double click. '''
    
        entry = self.disk[self.tree_store[path][3]]
        if not self.disk.mount(entry) :
            open_url(entry["FSTAB_PATH"])
        
    def on_enable_toggled(self, renderer, path) :
    
        use = self.tree_store[path][0]
        entry = self.disk[self.tree_store[path][3]]
        self.treeview2.set_cursor(path)
        if use :
            self.disk.unconfigure(entry)
        else :
            self.disk.configure(entry)
        
    def on_cursor_changed(self, treeview) :
        ''' Set button when we change device in the advance configuration '''

        path = self.treeview2.get_cursor()[0]
        logging.debug("on_cursor_changed: path = {}".format(path))
        if path is None:
            return
        try :
            idx = self.tree_store[path][3]
            entry = self.disk[idx]
        except NotInDatabase :
            self.info_button.set_sensitive(False)
            self.edit_button.set_sensitive(False)
            self.mount_button.set_sensitive(False)
            return
        sensitive = self.tree_store[path][1]
        self.info_button.set_sensitive(True)
        self.edit_button.set_sensitive(True)
        self.mount_button.set_sensitive(sensitive)
        if entry.get_is_mounted() :
            self.icon_mount.set_from_stock(Gtk.STOCK_CANCEL, Gtk.IconSize.BUTTON)
            self.label_mount.set_label(_("Unmount"))
        else :
            self.icon_mount.set_from_stock(Gtk.STOCK_APPLY, Gtk.IconSize.BUTTON)
            self.label_mount.set_label(_("Mount"))
            
    def on_revert_clicked(self, button) :
        ''' Come back to the original fstab when we click on revert '''
    
        ret = dialog("question", _("Reverting to an older version?"), \
            [_("This will apply the following changes:"), _("Do you want to continue?")],
            "\n".join(self.disk.get_changes_current_to_original()), parent = self.window_main)
        if ret[0] == Gtk.ResponseType.YES:
            self.disk.undo()
        
    def on_history_clicked(self, button) :
        ''' Start HistoryDialog. '''
    
        dial = HistoryDialog(self.disk, parent = self.window_main)
        dial.dialog_history.run()
        
    def on_save_clicked(self, button) :
    
        self.disk.savelog()
        self.update_main()
        
    def on_resize(self, widget, event) :
    
        self.width = event.width
        self.height = event.height
    
    def on_close_clicked(self, button) :
        ''' Save the log when we clo            AddWizard.py \
se the app '''

        self.window_main.hide()
        while Gtk.events_pending() :
            Gtk.main_iteration()
        self.conf.set("Gui Config", "selected", \
                self.tree_store.get_sort_column_id()[0])
        self.conf.set("Gui Config", "selected_order", \
                self.tree_store.get_sort_column_id()[1].value_nick)
        self.conf.set("Gui Config", "main_width", self.width)
        self.conf.set("Gui Config", "main_height", self.height)
        self.disk.shutdown()
        Gtk.main_quit()



# -*- coding: UTF-8 -*-
#
#  Fstab.py : Provide low level classes to manage fstab/mtab like file
#  Copyright (C) 2007 Mertens Florent <flomertens@gmail.com>
#  Updated 2021 for MX Linux Project by team member Nite Coder
#  Maintenance of project assumed by MX Linux with permission from original author.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import os
import re
import copy
import shutil
import tempfile
import functools
import subprocess
from subprocess import *

from . import FstabData
from .DiskInfo import *
from .FstabError import *
from .FstabUtility import *


class EntryBase(dict) :
    ''' EntryBase([device/uuid/label, path, option, type, freq, paso]) -> new EntryBase\n
        Create a dictionary of information available about device :
        - Merge information provided in argument with DiskInfo information.
        - Store all fstab/mtab related attribute as FSTAB_* '''

    def __init__(self, entry) :

        if "info" not in globals() :
            globals()["info"] = get_diskinfo_backend("auto")()
        for i in FstabData.categorie : self[i] = ""
        for i in range(len(entry)) : self[FstabData.categorie[i]] = entry[i]
        if self["FSTAB_PATH"] :
            self["FSTAB_PATH"] = decode(self["FSTAB_PATH"])
        try :
            self.update(info[info.search_device(entry[0])])
        except :
            pass

    def write(self, type = "FSTAB_NAME") :
        ''' w.write() -> Return a string of the entry in fstab/mtab synthax '''

        result = ""
        if type == "FS_UUID" or type == "FS_LABEL" and "DEVICE" in self :
            result += "#Entry for " + self["DEVICE"] + " :\n%s=" % type.split("_")[-1]
        elif type == "DEVICE" and "FSTAB_NAME" in self and os.path.exists(self["FSTAB_NAME"]) :
            if os.path.samefile(self["DEVICE"], self["FSTAB_NAME"]) :
                type = "FSTAB_NAME"
        result += "%s\t" % self[type]
        result += "%s\t" % encode(self["FSTAB_PATH"])
        result += "\t".join(self[k] for k in FstabData.categorie[2:])
        result += "\n"
        return result

class Entry(EntryBase) :
    ''' Entry([device/uuid/label, [path], [option], [type], [freq], [paso]]) -> new Entry\n
        Difference with EntryBase :
        - Only device is required. If not provided, other required info are set to defaults.
        - Set FSTAB_OPTION, FSTAB_FREQ and FSTAB_PASO to default when changing FSTAB_TYPE.
        - If FSTAB_PATH is in use, add an "_" until it can be used.
        - Set special option locale=autoset automatically. This possibility will be extend
          to other options in the future.
        - When setting FSTAB_TYPE to FS_TYPE, we use the first value of FS_DRIVERS, which may be
          different. example : FS_TYPE is ntfs and you don't have the ntfs driver, but ntfs-3g '''

    def __init__(self, entry, parent = None) :

        self.parent = parent
        EntryBase.__init__(self, entry)
        if not self["FSTAB_PATH"] :
            if "FS_LABEL_SAFE" in self :
                self["FSTAB_PATH"] = "/media/" + self["FS_LABEL_SAFE"]
            else :
                self["FSTAB_PATH"]  = "/media/" + self["DEV"]
        if not self["FSTAB_TYPE"] :
            self["FSTAB_TYPE"] = self["FS_TYPE"]

    def __setitem__(self, key, item) :
        ''' x.__setitem__(i, y) <==> x[i]=y '''

        if key == "FSTAB_TYPE" and item :
            if "FS_TYPE" in self and item == self["FS_TYPE"] :
                try :
                    item = self["FS_DRIVERS"]["primary"][0][0]
                except :
                    item = self["FS_TYPE"]
            if "FSTAB_OPTION" in self :
                for value in self.listopt():
                    if value not in FstabData.common :
                        self.removeopt(value)
                default = list(FstabData.defaults.get(item, FstabData.defaults["__default__"]))
                if self["FSTAB_OPTION"] :
                    default[0] += "," + self["FSTAB_OPTION"]
                (self["FSTAB_OPTION"], self["FSTAB_FREQ"], self["FSTAB_PASO"]) = default

        if key == "FSTAB_OPTION" and item and "FSTAB_OPTION" in self :
            listopt = item.split(",")
            if "locale=autoset" in listopt :
                listopt.remove("locale=autoset")
                listopt.append("locale=" + os.environ["LANG"])
                item = ",".join(listopt)

        if key == "FSTAB_PATH" and item and "DEVICE" in self \
                and "FSTAB_PATH" in self :
            while not check_path(item, self.parent, entry = self) :
                logging.debug("-> %s is in use." % item)
                item = item + "_"
            item = os.path.normpath(item)

        dict.__setitem__(self, key, item)

    def __eq__(x, y) :

        x2 = dict.copy(x)
        x2["FSTAB_NAME"] = y["FSTAB_NAME"]
        return dict.__eq__(x2, y)

    def make_path(self) :
        ''' x.make_path() -> Create FSTAB_PATH.\n
            If path is already in use add a "_" until it can be used.
            All path created by the class are stored in %s ''' % CREATED_PATH_FILE

        path = self["FSTAB_PATH"]
        logging.debug("Check if %s is created" % path)
        while not check_path(path, self.parent, entry = self) :
            logging.debug("-> %s is in use." % path)
            path = path + "_"
        for i in range(len(path.split("/")) -1) :
            if not os.path.exists("/".join(path.split("/")[:i+2])) :
                logging.debug("-> create path %s" % "/".join(path.split("/")[:i+2]))
                try :
                    os.mkdir("/".join(path.split("/")[:i+2]))
                except OSError:
                    logging.error("Creating %s failled. "
                       "There is something wrong with this path." % "/".join(path.split("/")[:i+2]))
                list_created_path("add", "/".join(path.split("/")[:i+2]))
        if not self["FSTAB_PATH"] == path :
            self["FSTAB_PATH"] = path

    def listopt(self) :
        ''' x.listopt() -> return a list of FSTAB_OPTION '''

        return self["FSTAB_OPTION"].split(",")

    def hasopt(self, option) :
        ''' x.hasopt(option) -> True/False '''

        if option in self.listopt() :
            return True
        return False

    def addopt(self, options) :
        ''' x.addopt([opt1,opt2 ...]) -> Add opt1, opt2 ... to FSTAB_OPTION '''

        if isinstance(options, str) :
            options = options.split(",")
        for value in options :
            if value not in self.listopt() :
                if self["FSTAB_OPTION"] :
                    self["FSTAB_OPTION"] += "," + value
                else :
                    self["FSTAB_OPTION"] = value

    def removeopt(self, options) :
        ''' x.removeopt([opt1,opt2 ...]) -> Remove opt1, opt2 ... from FSTAB_OPTION '''

        if isinstance(options, str) :
            options = options.split(",")
        for value in options :
            if value in self.listopt() :
                opt = self.listopt()
                opt.remove(value)
                self["FSTAB_OPTION"] = ",".join(opt)

    def defaultopt(self) :
        ''' x.defaultopt() -> return the default options for the entry '''

        return FstabData.defaults.get(self["FSTAB_TYPE"], FstabData.defaults["__default__"])[0]

    def setopt(self, options) :
        ''' x.setopt([opt1,opt2 ...]) -> set opt1, opt2 ... to FSTAB_OPTION '''

        self["FSTAB_OPTION"] = ""
        self.addopt(options)

    def mount(self) :
        ''' x.mount() -> Mount the entry and return (exit_code, stderr+stdout)\n
            The path is created if it dont exist yet '''
        self.make_path()
        cmd = "%s %s %s -t %s -o %s" % (MOUNT, self["DEVICE"], escape_special(self["FSTAB_PATH"]), \
                            self["FSTAB_TYPE"], self["FSTAB_OPTION"])
        process = Popen(cmd, stderr=STDOUT, stdout=PIPE, close_fds=True, shell=True, text=True)
        sts = process.wait()
        output = process.stdout.read()
        logging.debug("Mounting %s on %s :\n-> cmd : %s\n-> exit : %s\n-> output : %s" % \
                (self["DEVICE"], self["FSTAB_PATH"], cmd, sts , output))

        fstab_path = self["FSTAB_PATH"]
        chk_bind = "test ! -d /run/systemd/system && grep '" + fstab_path +"' /etc/fstab | awk '/^[a-zA-Z/]/{print $4}' | grep bind"
        try :
            logging.debug("TRY CHK_BIND: %s on %s :\n-> cmd : %s\n" % (self["DEVICE"], fstab_path, chk_bind))
            prc = run(chk_bind, shell=True, stderr=STDOUT, stdout=PIPE, check=True, text=True)
            ret = prc.returncode
            out = prc.stdout
            logging.debug("TRY CHK_BIND: %s" % out)
            try :
                logging.debug("MOUNT: bind mounts")
                prc = run([MOUNT, "-a"], stderr=STDOUT, stdout=PIPE, check=True, text=True)
                ret = prc.returncode
                out = prc.stdout
            except:
                pass
        except:
            pass
        return (sts, output)

    def umount(self, lazy = False) :
        ''' x.umount() -> Unmount the entry and return (exit_code, stderr+stdout)
            fix bind mounts by unmounting all mountpoints related to device '''

        device = self["DEVICE"]
        fstab_path = self["FSTAB_PATH"]

        # get mountpoints based on device as listed in /proc/mounts
        mountpoints = []
        with open('/proc/mounts', 'rt') as mounts:
            for mount in mounts:
                    [dev, mp] = mount.strip().split()[0:2]
                    if dev == device:
                        logging.debug(f"Found mountpoint of {dev} on {mp}")
                        mountpoints += [decode(mp)]
        ret = 0
        out = ""
        # unmount mountpoints in reverse order
        for mountpoint in mountpoints[::-1]:
            if not os.path.ismount(mountpoint):
                continue
            cmd = [ UMOUNT, "--recursive", mountpoint ]
            #cmd = [ UMOUNT, mountpoint ]
            if lazy :
                cmd += ["-l"]
            try:
                prc = run(cmd, stderr=STDOUT, stdout=PIPE, check=True, text=True)
                ret = prc.returncode
                out = prc.stdout
                logging.debug(f"Unmounting {device} on {mountpoint} :\n-> cmd : {' '.join(cmd)}\n-> exit : {ret}\n-> output : {out}")
            except CalledProcessError as e:
                ret = e.returncode
                out = e.stdout
                logging.debug(f"Unmounting {device} on {mountpoint} :\n-> cmd : {' '.join(cmd)}\n-> exit : {ret}\n-> output : {out}")
                break

        return (ret, out)

    def get_is_mounted(self) :
        ''' x.get_is_mounted() -> True/False '''

        return bool(os.path.ismount(self["FSTAB_PATH"]))

    def get_is_system(self) :
        ''' x.get_is_mounted() -> True/False\n
            System partition are FSTAB_PATH before /home/*, /media/* and /opt.
            So /home is sytem, but /opt is not. '''

        path = os.path.normpath(self["FSTAB_PATH"])
        if path in FstabData.system["exact"] + FstabData.system["extand"] :
            return True
        for sys_path in FstabData.system["extand"] :
            if path[:len(sys_path)+1] == "%s/" % sys_path :
                return True
        return False

    def get_size(self) :
        ''' x.get_size() -> Return size of device '''

        if self.get_is_mounted() :
            return os.statvfs(self["FSTAB_PATH"])[2]*os.statvfs(self["FSTAB_PATH"])[1]
        else :
            return self["SIZE"]

    def get_free_size(self) :
        ''' x.get_free_size() -> Return free size of device\n
            If device is not mounted, return 0 '''

        if self.get_is_mounted() :
            stat = os.statvfs(self["FSTAB_PATH"])
            return stat[3]*stat[1]
        else :
            return 0

    def get_available_size(self) :
        ''' x.get_available_size() -> Return available size of device\n
            If device is not mounted, return 0 '''

        if self.get_is_mounted() :
            stat = os.statvfs(self["FSTAB_PATH"])
            return stat[4]*stat[1]
        else :
            return 0

    def get_used_size(self) :
        ''' x.get_used_size() -> Return available size of device\n
            If device is not mounted, return 0 '''

        if self.get_is_mounted() :
            return self.get_size() - self.get_free_size()
        else :
            return 0

    def copy(self) :
        ''' x.copy() -> return an exact copy of the entry '''

        new = self.__class__([self["FSTAB_NAME"], self["FSTAB_PATH"], self["FSTAB_TYPE"], \
            self["FSTAB_OPTION"], self["FSTAB_FREQ"], self["FSTAB_PASO"]], parent = self.parent)
        return new

class MntFile(list) :
    ''' MntFile(filename, [fd], [minimal], [naming], [backend]) -> new MntFile object\n
        MntFile create a list of Entry object for each manageable entry of filename (or fd).
        An entry should be a line of 6 field separeted by at least one space :
        device/uuid/label, path, option, type, freq, and paso. (see man fstab)
        A manageable device is a none ignored device of DiskInfo dadatbase.
        Other 6 field entry are stored as EntryBase respectively in x.dev for virtual device,
        x.other for other device, x.comment for commented entry (ex : #device path ...).
        You can get an Entry from the MntFile, either by specify is index,
        just like with list, but also by specify it path.
        ex : x[i] -> Entry at index i of MntFile
             x[\"/\"] - > Entry with FSTAB_PATH = \"/\"
        If the Entry can't be found, the NotInDatabase exception will be raised
        Option available :
        - fd : create MntFile from fd
        - minimal : if True, create a minimal MntFile, with only EntryBase object.
          Faster to create and to manage. Default to False
        - naming : specify the naming that should be used when writing filename back
          available naming are : uuid, dev or auto. auto will use uuid if at least
          one entry of the file use uuid. Default to auto.
        - backend : specify the backend that should be use for DiskInfo. Default to auto '''

    def __init__(self, filename, fd = None, minimal = False, \
            naming = "auto", backend = "auto") :

        if "info" not in globals() :
            globals()["info"] = get_diskinfo_backend(backend)()
        self.info = info
        (self.dev, self.other, self.comment) = ([], [], [])
        self.filename = filename
        self.naming = naming
        if fd :
            mntfile = fd
        else :
            logging.debug(f"==>> Scanning {self.filename}")
            mntfile = open(self.filename)
        lines = mntfile.readlines()
        mntfile.close()
        if not minimal :
            logging.debug("Creating MntFile object for %s" % filename)
        seen = []
        for line in lines :
            entry = line.split()
            if len(entry) in [4,5] and not entry[0][0] == "#" :
                if len(entry) == 4 :
                    entry.append("0")
                if len(entry) == 5 :
                    entry.append("0")
            if not len(entry) == 6 :
                continue
            if self.filename == "/etc/mtab":
                if entry[0] in seen:
                    logging.debug(f"Skipping seen device in {filename}: {line.strip()}")
                    continue
                else:
                    seen += [ entry[0] ]

                #if entry[2] in "fuseblk":
                #    logging.debug(f"Skipping fuseblk mounts in {filename}: {line.strip()}")
                #    continue
                #else:
                #    seen += [ entry[0] ]

            if minimal :
                if not entry[0][0] == "#" :
                    self.append(EntryBase(entry))
                continue
            i = info.search_device(entry[0], ignored = False)
            if i :
                logging.debug("-> Matching %s on %s as %s" % (entry[0], entry[1], info.get(i, "DEV")))
                self.append(Entry(entry, self))
            elif entry[0] in FstabData.virtual_dev  :
                logging.debug("-> Matching %s on %s as virtaul fs" % (entry[0], entry[1]))
                self.dev.append(EntryBase(entry))
            elif not line.strip()[0] == "#" :
                logging.debug("-> Matching %s on %s as other fs" % (entry[0], entry[1]))
                self.other.append(EntryBase(entry))
            elif re.search(r"^#\S", entry[0]) :
                self.comment.append(EntryBase(entry))

    def __getitem__(self, item) :
        ''' x.__getitem__(y) <==> x[y] '''

        if isinstance(item, int) :
            try :
                return list.__getitem__(self, item)
            except :
                raise NotInDatabase(item)
        elif isinstance(item, slice) :
            try :
                return list.__getitem__(self, item)
            except :
                raise NotInDatabase(item)
        elif isinstance(item, str) :
            try :
                return list.__getitem__(self, self.search(item)[0])
            except :
                raise NotInDatabase(item)
        else :
            raise DatabaseTypeError(type(item), (int, str))

    def add(self, entry) :
        ''' x.add(entry) -> Create an Entry object with entry and add it to x.\n
            Return the new created entry '''

        self.append(Entry(entry, self))
        return self[-1]

    def remove(self, entry) :
        ''' x.remove(entry) -> remove entry from x\n
            entry should be the index of the Entry to delete a path name or the entry itself '''

        if isinstance(entry, int) :
            try :
                list.__delitem__(self, entry)
            except :
                raise NotInDatabase(entry)
        elif isinstance(entry, Entry) :
            try :
                list.remove(self, entry)
            except :
                raise NotInDatabase(entry)
        elif isinstance(entry, str) :
            try :
                list.__delitem__(self, self.search(entry, strict = "yes")[0])
            except :
                raise NotInDatabase(entry)
        else :
            raise DatabaseTypeError(type(entry), (int, str, Entry))

    def list(self, col = "DEVICE") :
        ''' x.list([col]) -> List all values of attribute col. Default to DEVICE '''

        result = []
        for k in self :
            if col in k :
                result.append(k[col])
            else :
                result.append("")
        return result

    def search(self, pattern, strict = "yes", keys = ["FSTAB_PATH"]) :
        ''' x.search(pattern, [strict], [list], [keys]) -> search for pattern in each
            keys of each Entry of x\n
            Default keys are : ["FSTAB_PATH"]
            If strict == "no", index entry where a substring pattern is found. Default to "yes". '''

        result = []
        for col in keys :
            i = 0
            for value in self.list(col) :
                if strict == "yes" :
                    if value == pattern and i not in result :
                        result.append(i)
                else :
                    if value.find(pattern) != -1 and i not in result :
                        result.append(i)
                i = i + 1
        return result

    def make_all_path(self) :
        ''' x.make_all_path() -> check that path used in MntFile are created, and create them if needed\n
            See make_path method of Entry object. '''

        logging.debug("Checking that all path used in %s are created :" % self.filename)
        for entry in self :
            entry.make_path()

    def write(self) :
        ''' x.write() -> Return a string of the MntFile in fstab/mtab default synthax '''

        result = ""
        for i in range(len(self.dev)) :
            result += self.dev[i].write()
        if not self.naming in ("auto", "dev", "uuid") :
            logging.warning("Fstab naming %s is not supported. Using auto naming." % self.naming)
            self.naming = "auto"
        if self.naming == "auto" and self.search("UUID=", strict = "no", keys = ["FSTAB_NAME"]) :
            self.naming = "uuid"
        self.sort(key=functools.cmp_to_key(self._sort_path))
        for entry in self + self.other :
            if "LABEL=" in entry["FSTAB_NAME"] and "FS_LABEL" in entry \
                    and len(self.search(entry["FS_LABEL"], keys = ["FS_LABEL"])) < 2 \
                    and os.path.exists("/dev/disk/by-label/%s" % entry["FS_LABEL"]) :
                type = "FS_LABEL"
            elif self.naming == "uuid" and "FS_UUID" in entry \
                    and len(self.search(entry["FS_UUID"], keys = ["FS_UUID"])) < 2 \
                    and os.path.exists("/dev/disk/by-uuid/%s" % entry["FS_UUID"]) :
                type = "FS_UUID"
            elif "DEVICE" in entry :
                type = "DEVICE"
            else :
                type = "FSTAB_NAME"
            result += entry.write(type)
        result += "\n"
        for i in range(len(self.comment)) :
            result += self.comment[i].write()
        result += "\n"
        return result

    def _sort_path(self, x, y) :

        x = x["FSTAB_PATH"]
        y = y["FSTAB_PATH"]
        if x == y :
            return 0
        if x.startswith(y) :
            return 1
        if y.startswith(x) :
            return -1
        ix = iy = 0
        for i, path in enumerate(FstabData.path_order) :
            if x.startswith(path) :
                ix = i
            if y.startswith(path) :
                iy = i
        return ix - iy

    def copy(self) :
        ''' x.copy() -> Create an exact copy of the MntFile '''

        new = copy.copy(self)
        for i in range(len(self)) :
                new[i] = self[i].copy()
        return new

    def apply(self) :
        ''' x.apply() -> Write MntFile to filename '''

        tmpfile = tempfile.NamedTemporaryFile()
        tmpfile.write(FstabData.header.encode('utf-8'))
        tmpfile.write(self.write().encode('utf-8'))
        tmpfile.seek(0)
        mntfile = open(self.filename, "wb")
        shutil.copyfileobj(tmpfile, mntfile)
        mntfile.close()
        tmpfile.close()


def list_created_path(action = "list", path = None) :
    ''' Action :
        "list" : list all path created by python-fstab
        "add"  : add a path to the list
        "del"  : remove a path from the list
        "clean": clean the file '''

    created_file = open(CREATED_PATH_FILE, "a+")
    li = created_file.read().strip().split("\n")
    created_file.close()
    created_file = open(CREATED_PATH_FILE, "w+")
    if action == "add" and not encode(path) in li :
        li.append(encode(path))
    if action == "del" and encode(path) in li :
        li.remove(encode(path))
    if action == "clean" :
        for path in li[:] :
            if not os.path.exists(decode(path)) :
                logging.debug("Clean %s from the list of created path" % path)
                li.remove(path)
    created_file.write("\n".join(li).strip())
    created_file.close()
    if not "".join(li) and os.path.exists(CREATED_PATH_FILE) :
        os.remove(CREATED_PATH_FILE)
    if action == "list" :
        if len(li) == 1 and len(li[0]) == 0 :
            return []
        return list(map(decode, li))

def clean_path(path) :
    ''' Remove the path if not in use '''

    logging.debug("Check if path can be deleted : %s" % path)
    while path in list_created_path() :
        if os.path.isdir(path) and check_path(path) or not os.path.exists(path) :
            logging.debug("-> delete path : %s" % path)
            try :
                os.rmdir(path)
                list_created_path("del", path)
            except OSError :
                logging.error("Deleting %s failled. "
                    "There is something wrong with this path." % path)
        path = os.path.dirname(path)

def clean_all_path() :
    ''' Check all path in list_created_path, and reomove them if not in use '''

    logging.debug("Checking all path for deletation :")
    for path in list_created_path() :
        clean_path(path)

def check_path(path, fstab = None, entry = None) :
    ''' check that path is not in use '''

    path = os.path.normpath(path)
    if os.path.ismount(path) :
        if entry and path == entry["FSTAB_PATH"] :
            return True
        return False
    else :
        not_in_use = bool(not os.path.exists(path) or ( os.path.isdir(path) and len(os.listdir(path)) == 0 ))
        if not fstab :
            fstab = MntFile(FSTAB, minimal = True)
        if fstab.search(path) :
            if entry and path == entry["FSTAB_PATH"] and not_in_use :
                return True
            return False
        if not_in_use :
            return True
        return False

def device_is_mounted(device) :
    ''' Return path used by device if device is mounted.
        This function differ from the get_is_mounted() method of Entry as this function
        try to find device in mtab, and don't just look if a device is mounted at path '''

    mtab = MntFile(MTAB, minimal = True)
    return [ mtab[i]["FSTAB_PATH"] for i in mtab.search(device, keys = ["DEV", "DEVICE"]) ]

if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG, format='%(levelname)s : %(message)s')
    fstab = MntFile("fstab.test", minimal=True)
    print(fstab.list("FSTAB_PATH"))
    fstab.sort(cmp=fstab._sort_path)
    print(fstab.list("FSTAB_PATH"))



# -*- coding: UTF-8 -*-
#
#  ToolsBackend.py : DiskInfo Backend that use various tools
#                    to get device informations
#  Copyright (C) 2007 Mertens Florent <flomertens@gmail.com>
#  Updated 2021 for MX Linux Project by team member Nite Coder
#  Maintenance of project assumed by MX Linux with permission from original author.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#


import os
import re
import logging
from subprocess import *

from Fstabconfig import *
from FstabError import *
import FstabData
from DiskInfo import DiskInfoBase

class ToolsBackend(DiskInfoBase) :

    def __init__(self) :
    
        DiskInfoBase.__init__(self)

    def _backend_score(self) :
        ''' Return the reliability score of the backend (0-100) '''
        
        return 50

    def _exec(self, cmd):
        
        logging.debug("Executing : %s" % cmd)
        process = Popen(cmd, stderr=STDOUT, stdout=PIPE, close_fds=True, shell=True)
        sts = process.wait()
        if not sts :
            res = "Success"
        else :
            res = "Failure"
        logging.debug("-> Exit status : %s (%s)" % (str(sts), res))
        return (sts, process.stdout.readlines())
        
    def _get_device_info(self, device) :
            
        dev = {}
        # Get dev, size, major and minor directly from /proc/partitions
        dev["DEV"] = device.strip().split()[3]
        logging.debug("Looking at : " + dev["DEV"])
        dev["SIZE"] = float(device.strip().split()[2])*1024.
        dev["MAJOR"] = device.strip().split()[0]
        dev["MINOR"] = device.strip().split()[1]
        
        # See if this partition is from a parent already in the database
        try :
            drive = re.search("^([a-z]{1,4}|(nvme[0-9]{1,2}n[0-9]{1,2}|mmcblk[0-9]{1,2})p)[0-9]{1,3}$", dev["DEV"]).groups()[0]
            if any(y in drive for y in ["mmcblk", "nvme"]):
                drive = drive[:-1]
            dev["PARENT"] = self[drive]["DEV"]
        except NotInDatabase :
            dev["PARENT"] = None
        except AttributeError :
            dev["PARENT"] = None
        logging.debug("-> Set   PARENT :" + str(dev["PARENT"]))
            
        # Set sysfs_path and /removable path depending of the previous result
        if dev["PARENT"] :
            dev["SYSFS_PATH"] = "/sys/block/" + dev["PARENT"] + "/" + dev["DEV"]
            removable = "/sys/block/" + dev["PARENT"] + "/removable"
        else :
            dev["SYSFS_PATH"] = "/sys/block/" + dev["DEV"]
            removable = "/sys/block/" + dev["DEV"] + "/removable"
            
        # Set removable
        if os.path.exists(removable) :
            fd = open(removable)
            dev["REMOVABLE"] = bool(int(fd.readline()))
        else :
            dev["REMOVABLE"] = False
        logging.debug("-> Set   REMOVABLE :" + str(dev["REMOVABLE"]))
        
        # Get the more info posible about the device from udevinfo
        cmd = UDEVINFO + " -p " + dev["SYSFS_PATH"].replace("/sys","") + " -q all"
        logging.debug("UDEVINFO : " + cmd)
        (sts, result) = self._exec(cmd)
        if not sts :
            for line in result :
                logging.debug("(udevinfo output) " + line.strip().decode("utf-8", errors="ignore"))
                try :
                    device = "/dev/" + re.search(r"N: (\S+)", line.decode("utf-8", errors="ignore")).groups()[0]
                    dev["DEVICE"] = device
                    logging.debug("-> Found DEVICE : " + device)
                except AttributeError :
                    pass
                # Used for debugging blkid & vol_id :
                # continue
                try :
                    (attr, value) = re.search(r"E: ID_(\S+)=(.+)", line.decode('utf-8', errors="ignore")).groups()
                    dev[attr] = value
                    logging.debug("-> Found " + attr + " : "+ value)
                except AttributeError :
                    pass
                
        # Used for debugging dmsetup :
        #if not dev["DEV"].find("dm-") == -1 and dev.has_key("DEVICE") :
        #    del dev["DEVICE"]
                
        # If udevinfo fail or didn't give DEVICE, it's might be a mapped device.
        # Try to get DEVICE from dmsetup
        if "DEVICE" not in dev and "dm-" in dev["DEV"] :
            cmd = DMSETUP + " info -j " + dev["MAJOR"] + " -m " + dev["MINOR"] + " | grep Name"
            (sts, result) = self._exec(cmd)
            result = "".join(result)
            logging.debug("(dmsetup output) " + result.strip())
            if not sts and result :
                device = "/dev/mapper/" + result.split()[-1]
                logging.debug("-> Found DEVICE : " + device)
                if os.path.exists(device) :
                    logging.debug("-> Check DEVICE : ok")
                    dev["DEVICE"] = device
                else :
                    logging.debug("-> Check DEVICE : fail")
            else :
                logging.debug("Warning : First attempt failed. Trying something else...")
                cmd = "%s ls |grep '(%s, %s)'" % (DMSETUP, dev["MAJOR"], dev["MINOR"])
                (sts, result) = self._exec(cmd)
                result = "".join(result)
                logging.debug("(dmsetup output) " + result.strip())
                if not sts and result :
                    device = "/dev/mapper/" + result.split()[0]
                    logging.debug("-> Found DEVICE : " + device)
                    if os.path.exists(device) :
                        logging.debug("-> Check DEVICE : ok")
                        dev["DEVICE"] = device
                    else :
                        logging.debug("-> Check DEVICE : fail")
        # If still no DEVICE, try /dev/ + DEV
        if "DEVICE" not in dev :
            device = "/dev/" + dev["DEV"]
            if os.path.exists(device) :
                logging.debug("Warning : no DEVICE found, try " + device)
                stat = os.stat(device)
                if stat.st_rdev == os.makedev(int(dev["MAJOR"]), int(dev["MINOR"])) :
                    logging.debug("-> Check DEVICE : ok")
                    logging.debug("-> Set   DEVICE : " + device)
                    dev["DEVICE"] = device
                else :
                    logging.debug("-> Check DEVICE : fail")
                                
        # Try to get slaves, set removable depending of the slave, and set FS_USAGE to other
        # if it is not already set. Here we don't really care what kind of type it is.
        # If it is manageable by the system, udeinfo or vol_id should have given us it real FS_USAGE.
        if os.path.isdir(dev["SYSFS_PATH"] + "/slaves") and os.listdir(dev["SYSFS_PATH"] + "/slaves") :
            dev["SLAVES"] = os.listdir(dev["SYSFS_PATH"] + "/slaves")
            dev["REMOVABLE"] = True
            for slave in dev["SLAVES"] :
                if self.search(slave) :
                    logging.debug("-> Add   SLAVES : " + slave)
                    if "FS_USAGE" not in self[slave] or self[slave] == "filesystem" :
                        self[slave]["FS_USAGE"] = "other"
                        logging.debug("-> Set   %s FS_USAGE : other" % slave)
                    dev["REMOVABLE"] = dev["REMOVABLE"] and self[slave]["REMOVABLE"] 
            logging.debug("-> Set   REMOVABLE : " + str(dev["REMOVABLE"]) )
        
        # Call vol_id to complete informations. This is important even if udevinfo
        # worked successfully, because udev db might be not up to date, and udev
        # might be wrong on some FS_LABEL (ones with " for exemple)    
        if "DEVICE" in dev :
            if "FS_USAGE" not in dev :
                cmd = BLKID + " " + dev["DEVICE"]
                (sts, result) = self._exec(cmd)
                result = "".join([i.decode('utf-8') for i in result])
                if not sts :
                    logging.debug("(blkid output) " + result.strip())
                    for attr in ("TYPE", "LABEL", "UUID") :
                        try :
                            pattern = "\\b" + attr + "=\"([^\"]+)\""
                            value = re.search(pattern, result).groups()[0]
                            if "FS_" + attr not in dev :
                                dev["FS_" + attr] = re.search(pattern, result).groups()[0]
                                logging.debug("-> Found " + attr + " : " + dev["FS_" + attr])
                            if attr == "TYPE" :
                                if value not in FstabData.ignore_fs :
                                    dev["FS_USAGE"] = "filesystem"
                                else :
                                    dev["FS_USAGE"] = "other"
                                logging.debug("-> Set   FS_USAGE : " + dev["FS_USAGE"])
                        except AttributeError : 
                            pass
                    
        for attr in ("UUID", "LABEL") :
            # If no UUID, LABEL try to get them manually
            if "DEVICE" in dev and "FS_" + attr not in dev :
                if os.path.exists(dev["DEVICE"]) and os.path.isdir("/dev/disk/by-" + attr.lower()) :
                    for file in os.listdir("/dev/disk/by-" + attr.lower()) :
                        if os.path.samefile("/dev/disk/by-" + attr.lower() + "/" + file, dev["DEVICE"]) :
                            logging.debug("-> Set   " + attr + " : " + file)
                            dev["FS_" + attr] = file

        return dev




# -*- coding: UTF-8 -*-
#
#  EventHandler.py : Execute actions on events detection
#  Copyright (C) 2007 Mertens Florent <flomertens@gmail.com>
#  Updated 2021 for MX Linux Project by team member Nite Coder
#  Maintenance of project assumed by MX Linux with permission from original author.

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import os
import time
import logging
import threading

from .Fstabconfig import *
from .Fstab import *
from .FstabError import *

class EventHandler :

    def __init__(self, disk, enable_watch) :

        self.disk = disk
        self.enable_watch = enable_watch
        self.filename = self.disk.filename
        self.inhibit =  False
        self.start_watch()
        
    def start_watch(self) :

        if not self.enable_watch :
            return
        logging.debug("Starting Watch...")
        self.th_watch = threading.Thread(target=self.start_watch_thread, name="WatchThread")
        self.th_watch.setDaemon(True)
        self.th_watch.start()
                
    def start_watch_thread(self) :
    
        self.alive = True
        self.mtime_mtab = os.stat(MTAB).st_mtime
        self.mtime_fstab = os.stat(self.filename).st_mtime
        self.size = [ str(k.get_free_size())[:4] for k in self.disk ]
        while self.alive :
            time.sleep(0.5)
            if not self.inhibit and not self.mtime_fstab == os.stat(self.filename).st_mtime :
                logging.debug("FSTAB update detected")
                self.emit("external_fstab_changed")
            if not self.inhibit and not self.mtime_mtab == os.stat(MTAB).st_mtime :
                logging.debug("MTAB update detected")
                self.emit("external_mtab_changed")  
            if not self.inhibit and not self.size == [ str(k.get_free_size())[:4] for k in self.disk ] :
                logging.debug("Size update detected")
                self.emit("size_changed")
                
    def stop_watch(self) :

        if hasattr(self, "th_watch") and self.th_watch.is_alive() :
            self.alive = False
            logging.debug("Stopping Watch...")
            self.th_watch.join()

    def shutdown(self) :
        
        t = time.time()
        self.stop_watch()
        logging.debug("EventHandler shutdown in %s s", time.time() -t)
        
    def emit(self, event) :
    
        if not hasattr(self, "on_%s" % event) :
            raise UnknowEvent(self, event)
        t = time.time()
        logging.debug("Emitting '%s' event" % event)
        if getattr(self, "on_%s" % event)() :
            if hasattr(self, "%s_fct" % event) :
                fct = getattr(self, "%s_fct" % event)
                if callable(fct[0]) :
                    logging.debug("-> Call %s%s for event '%s'" % (fct[0].__name__, fct[1], event))
                    fct[0](*fct[1])
            else :
                logging.debug("-> No function define for event '%s'" % event)
        logging.debug("Event '%s' managed in %s s", event, time.time() -t)

    def on_external_fstab_changed(self) :
    
        fstab = MntFile(self.filename)
        if self.disk.fstab[:] == fstab[:] :
            self.mtime_fstab = os.stat(self.filename).st_mtime
            logging.debug("In fact, no real changes.")
            return False
        self.disk._rebuild_object(fstab)
        self.disk._current = self.disk.copy()
        self.disk._current.fstab = self.disk.fstab.copy()
        self.emit("external_changed")
        self.mtime_fstab = os.stat(self.filename).st_mtime
        
    def on_external_mtab_changed(self) :
    
        mtab = MntFile(MTAB, minimal = True)
        self.disk._build_object(self.disk.fstab)
        self.disk._current = self.disk.copy()
        self.disk._current.fstab = self.disk.fstab.copy()
        self.emit("external_changed")
        self.mtime_mtab = os.stat(MTAB).st_mtime
        self.size = [ str(k.get_free_size())[:4] for k in self.disk ]
        
    def on_internal_changed_prepare(self) :

        logging.debug("Inhibit watch for internal change...")
        self.inhibit = True
        
    def on_internal_changed_cancel(self) :

        logging.debug("Restart watch...")
        self.inhibit = False
        
    def on_internal_changed(self) :
    
        self.emit("configuration_changed")
        self.mtime_mtab = os.stat(MTAB).st_mtime
        self.mtime_fstab = os.stat(self.filename).st_mtime
        self.size = [ str(k.get_free_size())[:4] for k in self.disk ]
        logging.debug("Restart watch...")
        self.inhibit = False
        
    def on_external_changed(self) :
    
        self.emit("configuration_changed")
        
    def on_configuration_changed(self) :
    
        self.emit("any_changed")
    
    def on_size_changed(self) :
        
        self.emit("any_changed")
        self.size = [ str(k.get_free_size())[:4] for k in self.disk ]
            
    def on_any_changed(self) :
    
        return True




# -*- coding: UTF-8 -*-
#
#  Utility.py : Various utility for Fstab
#  Copyright (C) 2007 Mertens Florent <flomertens@gmail.com>
#  Updated 2021 for MX Linux Project by team member Nite Coder
#  Maintenance of project assumed by MX Linux with permission from original author.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import os
import sys
import time
import glob
from subprocess import *

from . import FstabData
from .Fstabconfig import *

def get_used_file(path) :

    cmd = "%s +c 0 %s" %(LSOF, escape_special(path))
    lines = getoutput(cmd).split("\n")
    result = []
    for line in lines :
        used = path + line.split(path)[-1].strip()
        if ( used == path or not used.find(path + "/") == -1 ) and len(line.split()) > 0 :
            result.append(line.split()[0] + " -> " + used)
    return result
  
def get_dmesg_output() :

    return "\n".join(getoutput(DMESG + " | tail").split("\n"))
    
def get_fuse_options(entry) :

    cmd = "ps -eo cmd | grep -F %s" % escape_special(entry["FSTAB_PATH"])
    lines = getoutput(cmd).split("\n")
    for line in lines :
        res = line.split()
        if os.path.exists(res[1]) and os.path.realpath(res[1]) == entry["DEVICE"] :
            entry["FSTAB_TYPE"] = entry["FS_TYPE"]
            for driver in glob.glob("/sbin/mount.*") :
                if os.path.basename(res[0]) == os.path.basename(driver) \
                        or os.path.basename(res[0]) == os.path.basename(os.path.realpath(driver)) :
                    entry["FSTAB_TYPE"] = driver.split("mount.")[-1]
            if "-o" in res :
                entry["FSTAB_OPTION"] = res[res.index("-o") + 1]

def encode(string) :
    ''' encode space to \040 si it can be use in fstab '''

    return string.replace(" ", "\\040")
    
def decode(string) :
    ''' decode \040 as space '''
    
    return string.replace("\\040", " ")
    
def escape_special(text) :
    ''' escape special charachter so it can be use in a command line '''

    for char in FstabData.special_char :
        text = text.replace(char, "\\%s" % char)
    return text

def change_tracker(x, y) :
    ''' Track changes between two version of a MntFile '''
        
    (log, to_mount, to_umount, removed) = ([], [], [], [])
    if not x == y :
        for i in range(len(y)) :
        
            iy = y[i]
            iydev = iy["DEVICE"]
            iypath = iy["FSTAB_PATH"]
            if not len(x.search(iydev, keys = ["DEVICE"])) == len(y.search(iydev, keys = ["DEVICE"])) :
                if not x.search(iypath) :
                    log.append("Adding %s on %s" % (iydev, iypath))
                    if not iy.get_is_mounted() :
                        to_mount.append(iy)
                continue
            
            if len(y.search(iydev, keys = ["DEVICE"])) > 1 :
                if x.search(iypath) :
                    ix = x[x.search(iypath)[0]]
                    ixpath = ix["FSTAB_PATH"]
                else :
                   log.append("Adding duplicate %s on %s" % (iydev, iypath))
                   if not iy.get_is_mounted() :
                        to_mount.append(iy)
                   continue
            else :
                ix = x[x.search(iydev, keys = ["DEVICE"])[0]]
                ixpath = ix["FSTAB_PATH"]
            
            if ix == iy :
                continue

            log.append("Changing %s on %s:" % (iydev, ixpath))
            if not iypath == ixpath :
                log.append("+ path %s -> %s" % (ixpath, iypath))
                if not iy.get_is_mounted() and ix.get_is_mounted() :
                    to_umount.append(ix)
                    to_mount.append(iy)

            if not iy["FSTAB_TYPE"] == ix["FSTAB_TYPE"] :
                log.append("+ type: %s -> %s" % (ix["FSTAB_TYPE"], iy["FSTAB_TYPE"]))
                if not iy in to_mount and ix.get_is_mounted() :
                    to_umount.append(ix)
                    to_mount.append(iy)

            if not iy["FSTAB_OPTION"] == ix["FSTAB_OPTION"] :
                log.append("+ options: %s -> %s" % (ix["FSTAB_OPTION"], iy["FSTAB_OPTION"]))
                if not iy in to_mount and ix.get_is_mounted() :
                    for opt in ix.listopt() :
                        if opt not in iy.listopt()  \
                                and opt not in FstabData.dont_need_remount :
                            to_umount.append(ix)
                            to_mount.append(iy)
                            break
                    for opt in iy.listopt() :
                        if opt not in ix.listopt() \
                                and opt not in FstabData.dont_need_remount :
                            to_umount.append(ix)
                            to_mount.append(iy)
                            break
                
            if not iy["FSTAB_PASO"] == ix["FSTAB_PASO"] :
                log.append("+ check at boot: %s -> %s" % (ix["FSTAB_PASO"], iy["FSTAB_PASO"]))

        for i in range(len(x)) :           
            ix = x[i]
            ixdev = ix["DEVICE"]
            ixpath = ix["FSTAB_PATH"]
            if not y.search(ixpath) :
                if not len(y.search(ixdev, keys = ["DEVICE"])) == len(x.search(ixdev, keys = ["DEVICE"])) :
                    log.append("Removing %s on %s" % (ixdev, ixpath))
                    if ix.get_is_mounted() :
                        removed.append(ix)
                elif not len(x.search(ixdev, keys = ["DEVICE"])) == 1 :
                    log.append("Removing duplicate %s on %s" % (ixdev, ixpath))
                    if ix.get_is_mounted() :
                        to_umount.append(ix)

    return (log, to_mount, to_umount, removed)

